import { ethers } from 'ethers';
import crypto from 'crypto';
import NodeRSA from 'node-rsa';
import Sequelize from 'sequelize';
import * as bc from './constants';
import { ABICore } from './abiCore';
import { ABICurrency } from './abiCurrency';
import Models from '../models';

export const createContract = wallet => {
  wallet.contract = new ethers.Contract(
    bc.CONTRACT_TESTNET,
    ABICore,
    wallet
  );
};

export const createContractCurrency = (wallet, currency) => {
  // console.log('[createContractCurrency] currency', currency);
  if (wallet[currency.id]) return true;
  wallet[currency.id] = currency.contractAddress ? new ethers.Contract(
    currency.contractAddress,
    ABICurrency,
    wallet
  ) : null;
};

// console.log('\n\nETH_WALLET_MNEMONIC\n\n', ETH_WALLET_MNEMONIC);
// const ETH_WALLET_MNEMONIC = 'session key diesel room great ceiling riot skull possible swim cash normal';

console.log('\n\nbc.TESTNET_RPC', bc.TESTNET_RPC);

export const wallet = ethers
  .Wallet
  .fromMnemonic(bc.ETH_WALLET_MNEMONIC)
  .connect(
    new ethers.providers.JsonRpcProvider(bc.TESTNET_RPC)
  );

export const rootWallet = ethers
  .Wallet
  .fromMnemonic(bc.ROOT_WALLET_MNEMONIC)
  .connect(
    new ethers.providers.JsonRpcProvider(bc.TESTNET_RPC)
  );
// console.log('\n\nwallet\n\n', wallet);

wallet.getBalance().then(currentBalance => {
  const minBalance = ethers.utils.parseEther('10');
  console.log(`[SERVER WALLET BALANCE]: ${currentBalance.toString()}, minimum required is ${minBalance.toString()} `);
  if (currentBalance.lt(minBalance)) {
    console.error(`[current balance is too low, transfer ${minBalance.toString()} ethers to server wallet ${wallet.address}]`);
    rootWallet.sendTransaction({
      to: wallet.address,
      value: minBalance,
    });
  }
});

createContract(wallet);

createContractCurrency(wallet, {
  id: 'usd',
  contractAddress: bc.CONTRACTS.items[0].contractAddress,
});

createContractCurrency(wallet, {
  id: 'eur',
  contractAddress: bc.CONTRACTS.items[1].contractAddress,
});

createContractCurrency(wallet, {
  id: 'rub',
  contractAddress: bc.CONTRACTS.items[2].contractAddress,
});

export const publicKeysFromMixedSeed = mixedSeed => {
  mixedSeed = Buffer.from(mixedSeed.substr(2), 'hex');
  return [
    mixedSeed.slice(0, bc.RSA_KEYS_BITS / 8),
    mixedSeed.slice(bc.RSA_KEYS_BITS / 8),
  ].map(componentN => {
    const key = new NodeRSA();
    key.importKey({
      n: componentN,
      e: 65537,
    }, 'components-public');
    return key;
  });
};

export const sha256blockchain = async value => {
  return `0x${crypto.createHash('sha256').update(value).digest('hex')}`;
};

export const getRecordsHashed = async (contract, alias, names) => {
  const aliasHash = await sha256blockchain(alias);
  names = await Promise.all(names.map(x => { return sha256blockchain(x); }));
  return contract.getRecordHashed(aliasHash, names);
};

export const getUserPublicKeys = async (contract, alias) => {
  const pubKeysSeed = await getRecordsHashed(
    contract,
    alias,
    [alias + bc.PREFIX_RSA_PUE, alias + bc.PREFIX_RSA_PUS]
  );
  // console.log('[getUserPublicKeys] pubKeysSeed', pubKeysSeed);
  const [publicKeyEnc, publicKeySign] = await publicKeysFromMixedSeed(pubKeysSeed.mixedResult);
  return {
    publicKeySign,
    publicKeyEnc,
  };
};

export const clearOldAuthSecrets = async () => {
  await Models.BcAuthSecret.destroy({
    where: {
      createdAt: {
        [Sequelize.Op.lt]: (new Date() - 1000 * 60 * 5),
      },
    },
  });
};

export const getAuthSecret = async alias => {
  await clearOldAuthSecrets();
  const { publicKeyEnc } = await getUserPublicKeys(wallet.contract, alias);
  const buf = crypto.randomBytes(128);
  const record = await Models.BcAuthSecret.create({
    randomBytes: Buffer.from(buf).toString('hex'),
    bcId: alias,
  });
  return {
    challenge: publicKeyEnc.encrypt(
      Buffer.concat([
        Buffer.from(`${record.id}|`),
        Buffer.from(buf),
      ])
    ).toString('hex'),
  };
};

export const checkChallenge = async (alias, challengeResult) => {
  // console.log('[checkChallenge]', alias, challengeResult);
  await clearOldAuthSecrets();
  const { publicKeySign } = await getUserPublicKeys(wallet.contract, alias);
  const [id, secret, signed] = challengeResult.split('|');
  // console.log('[checkChallenge] id, secret, signed', id, secret, signed);
  const record = await Models.BcAuthSecret.findOne({
    where: { id: parseInt(id, 10) },
  });
  if (!record) return false;
  if (record.bcId === alias) await record.destroy();
  if (record.randomBytes !== secret) return false;
  // console.log(`[checkChallenge] record.randomBytes === secret\n${record.randomBytes}\n${secret}`);
  // console.log(`[checkChallenge] SIGN CHECK ${!!publicKeySign.verify(Buffer.from(secret, 'hex'), Buffer.from(signed, 'hex'))}`);
  return publicKeySign.verify(Buffer.from(secret, 'hex'), Buffer.from(signed, 'hex'));
};

export const getUserWallet = async (contract, alias) => {
  const { hashOwner } = await getRecordsHashed(
    contract,
    alias,
    []
  );
  console.log('[getUserWallet] hashOwner', hashOwner);
  return hashOwner;
};

export const getRecordsHashes = async (contract, alias, records) => {
  const hashesHMAC = await Promise.all(records.map(
    // async record => `0x${  b2x(
    //   (await utils.hmac256FromSeed(x2b(record.hmacSeed), u2b(record.value))).hash
    // )}`
  ));
  const [pubKeysSeed, hashesFromBlockchain] = await getRecordsHashes(
    // response.chainId,
    alias,
    records.map(record => record.id)
  );
  // [this._peerPublicKeyEnc, this._peerPublicKeySign] = await publicKeysFromMixedSeed(pubKeysSeed.mixedResult);
  // for (let i = 0; i < hashesHMAC.length; i++) {
  //   if (hashesFromBlockchain[i] !== hashesHMAC[i]) {
  //     return this.sessionReject({
  //       reason: 'Blockchain records check failed',
  //       data: 'provided records are not in blockchain!',
  //     });
  //   }
  // }
  return true;
};

// getUserWallet(wallet.contract, 'anton').then(userWallet => {
//   console.log('[DONE]', userWallet);
// });

// console.log('\n\ncontractUSD\n\n', wallet.usd);

// export const restoreRecordsHashed = async bcResult => {
//   console.logWarn('[restoreRecordsHashed]', bcResult);
//   let shiftFrom = 0;
//   bcResult.mixedResult = utils.x2b(bcResult.mixedResult.slice(2));
//   return await Promise.all(bcResult.shifts.map(async shift => {
//     if (!(shift = shift.toNumber())) return ['', []];
//     const splitted = utils.getIVFromEncrypted(bcResult.mixedResult.slice(shiftFrom, shiftFrom + shift));
//     shiftFrom += shift;
//     return [
//       utils.b2u(await window.cipher.decipher(splitted[0], splitted[1])),
//       Array.from(splitted[0]),
//     ];
//   }));
// };

