import moment from 'moment';
import sharp from 'sharp';
import axios from 'axios/index';
import crypto from 'crypto';
import fs from 'fs';
import { ValidationError, AuthenticationError } from 'apollo-server';
import jwt from 'jsonwebtoken';
import * as globals from './globals';
import * as common from './constants';
import { isAdminUser } from '../resolvers/authorization';

export const getRandomIntInRange = (min, max) => {
  min = Math.ceil(min);
  max = Math.floor(max);
  return Math.floor(Math.random() * (max - min + 1)) + min;
};

export const createTimeInRange = (minCount, maxCount, hoursBefore, fromDate) => {
  const randomActions = getRandomIntInRange(minCount, maxCount);
  const timeNow = fromDate.getTime();
  return Array.from({ length: randomActions }).map(() => {
    return new Date(timeNow - ((Math.round(Math.random() * hoursBefore * 1000 * 60 * 60))));
  }).sort();
};

export const resizeImage = (buffer, width, height) => {
  return sharp(buffer).resize(width, height).toBuffer();
};

export const streamToBuffer = stream => {
  return new Promise((resolve, reject) => {
    const buffers = [];
    stream.on('data', buffer => {
      buffers.push(buffer);
    });
    stream.on('end', () => {
      resolve(Buffer.concat(buffers));
    });
    stream.on('close', () => {
      reject(new Error('Stream closed'));
    });
    stream.on('error', () => {
      reject(new Error('Stream error'));
    });
  });
};

export const utcTimeIsZero = dbTime => {
  return !moment.utc(dbTime).unix();
};

export const getUnixTimeStamp = dbTime => {
  return moment.utc(dbTime).unix();
};

export const nanoTime = () => {
  const hrTime = process.hrtime();
  return hrTime[0] * 1000000000 + hrTime[1];
};

export const createImagePath = uuid => {
  let res = '';
  uuid = uuid.replace('-', '');
  for (let i = 0; i < uuid.length; i++) {
    res += `${uuid[i]}/`;
    if (i > 6) {
      return {
        path: res,
        name: uuid.substr(i + 1),
      };
    }
  }
};

export const uploadAWS = async files => {
  return Promise.all(
    files.map(file => {
      return globals.S3.upload({
        Bucket: common.AWS_BUCKET_NAME,
        Body: file.buffer,
        Key: file.path,
      }).promise();
    })
  );
};

export const processAWSThumb = async ({
  imageBuffer, sessionId, prefix = '', onUpdateAction,
}) => {
  const imageStore = createImagePath(sessionId);
  const storePath = `${imageStore.path}${imageStore.name}`;
  const imagesWithNames = [];
  try {
    imagesWithNames.push({
      path: `${prefix}${storePath}_thumb_200x200.jpg`,
      buffer: await resizeImage(imageBuffer, 200, 200),
    });
    imagesWithNames.push({
      path: `${prefix}${storePath}_thumb_50x50.jpg`,
      buffer: await resizeImage(imageBuffer, 50, 50),
    });
    await uploadAWS(imagesWithNames);
  } catch (error) {
    console.log('--- processAWSThumb FAILED [bad images]');
  }
  onUpdateAction && onUpdateAction();
};

export const processAWS = async ({
  images,
  sessionId,
  prefix = '',
}) => {
  const imageStore = createImagePath(sessionId);
  const storePath = `${imageStore.path}${imageStore.name}`;
  const imagesWithNames = images.map((image, index) => ({
    path: `${prefix || ''}${storePath}_${index}.jpg`,
    buffer: image,
  }));
  try {
    await uploadAWS(imagesWithNames);
    return true;
  } catch (error) {
    return false;
  }
};

export const md5 = string => {
  return crypto.createHash('md5').update(string).digest('hex');
};

export const getRandomBytes = (length = 256) => {
  return new Promise(resolve => {
    crypto.randomBytes(length, (error, buffer) => {
      resolve(buffer);
    });
  });
};

export const string2uuid = string => {
  let result = `${string.substr(0, 8)}-${string.substr(8, string.length)}`;
  result = `${result.substr(0, 13)}-${result.substr(13, result.length)}`;
  result = `${result.substr(0, 18)}-${result.substr(18, result.length)}`;
  result = `${result.substr(0, 23)}-${result.substr(23, result.length)}`;
  return result;
};

export const isUuid = string => {
  const uuidRegex = /^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$/i;
  return uuidRegex.test(string);
};

export const s2b = s => {
  const b = new Uint8Array(s.length);
  for (let i = 0; i < s.length; i++) {
    const c = s.charCodeAt(i);
    if (c > 255) throw new Error('Wide characters are not allowed');
    b[i] = c;
  }
  return b;
};

export const a2b = s => {
  s += '===';
  s = s.slice(0, -s.length % 4);
  return s2b(atob(s.replace(/-/g, '+').replace(/_/g, '/')));
};

export const b2s = ab => {
  const b = (ab instanceof ArrayBuffer) ? new Uint8Array(ab) : ab;
  let s = '';
  for (let i = 0; i < b.length; i++) s += String.fromCharCode(b[i]);
  return s;
};

export const checkApiKey = async (uuid, models, me) => {
  const dbApiKey = await models.ApiKey.findByUUID(uuid);
  if (!dbApiKey) {
    throw new AuthenticationError(
      `Access violation [Wrong key ${uuid}]`
    );
  }
  if (dbApiKey.state === common.KEY_STATE_DEACTIVATED) {
    throw new AuthenticationError(
      `Access violation [Key is not active ${uuid}]`
    );
  }
  if (
    moment.utc(dbApiKey.expired).isBefore(moment())
    && !utcTimeIsZero(dbApiKey.expired)
  ) {
    throw new AuthenticationError(
      `Access violation [Key is expired ${uuid}]`
    );
  }
  const user = me || await models.User.findById(dbApiKey.userId);
  if (!user) {
    throw new AuthenticationError(
      'Access violation [Key owner not found]'
    );
  }
  return { dbApiKey, user };
};

export const checkApiKeyById = async (id, models, me) => {
  const dbApiKey = await models.ApiKey.findById(id);
  if (!dbApiKey) {
    throw new AuthenticationError(
      `Access violation [Wrong key id: ${id}]`
    );
  }
  if (dbApiKey.state === common.KEY_STATE_DEACTIVATED) {
    throw new AuthenticationError(
      `Access violation [Key [${id}] is not active]`
    );
  }
  if (
    moment.utc(dbApiKey.expired).isBefore(moment())
    && !utcTimeIsZero(dbApiKey.expired)
  ) {
    throw new AuthenticationError(
      `Access violation [Key [${id}] is expired]`
    );
  }
  const user = me || await models.User.findById(dbApiKey.userId);
  if (!user) {
    throw new AuthenticationError(
      'Access violation [Key owner not found]'
    );
  }
  return { dbApiKey, user };
};

export const getFiles = async (files, withDebug = true) => {
  // console.log('[getFiles] files', files);
  /* ********************* DEBUG *************** */
  if (withDebug && globals.DEBUG) {
    const totalImages = 21;
    const num = Math.floor(Math.random() * (totalImages));
    for (let i = 0; i < files.length; i++) {
      // files[i] = fs.readFileSync(`${__dirname}/../../_assets/test_images/0.jpg`);
      // files[i] = fs.readFileSync(`${__dirname}/../../_assets/test_images/vvp.jpg`);
      // files[i] = fs.readFileSync(`${__dirname}/../../_assets/test_images/${num}.jpg`);
      files[i] = fs.readFileSync(`${__dirname}/../../_assets/test_images/0.jpg`);
      // console.log('[getFiles] files[i]', files[i].byteLength);
    }
    return files;
  }
  /* ///////////////////// DEBUG /////////////// */

  return Promise.all(
    files.map(async file => {
      // console.log('[getFiles] file', file);
      const { createReadStream /* , filename, mimetype */ } = await file;
      // console.log('[getFiles] createReadStream', createReadStream);
      return streamToBuffer(createReadStream());
    })
  );
};

export const checkFiles = (files, count) => {
  if (files.length !== count) {
    throw new ValidationError(
      'Wrong images count for this service'
    );
  }
};

export const updateApiKey = async apiKey => {
  await apiKey.update({
    actionsLeft: apiKey.actionsLeft - 1,
  });
  // pubsub.publish(EVENTS.API_KEY.UPDATED, {
  //   apiKeyUpdated: { apiKey },
  //   userId: apiKey.userId,
  // });
};

export const createToken = async (user, secret, expiresIn) => {
  const {
    id, email, username, role,
  } = user;
  return jwt.sign({
    id, email, username, role,
  }, secret, {
    expiresIn,
  });
};

export const createStatsToken = async (user, secret) => {
  return jwt.sign({
    u: `${user.id}|${user.role}`,
  }, secret, {
    noTimestamp: true,
  });
};

export const validateStatsToken = async (token, secret) => {
  try {
    const statsToken = await jwt.verify(token, secret, { ignoreExpiration: true });
    const [userId, role] = statsToken.u.split('|');
    return {
      id: parseInt(userId, 10),
      role,
    };
  } catch (e) {
    throw new AuthenticationError(
      'Access denied [wrong token]'
    );
  }
};

export const logAction = (action, ...args) => {
  console.log(`
    ${common.LOG_SEP_PLUS}
      ${new Date()}
      ${action}
      ${args}
    ${common.LOG_SEP_PLUS}
  `);
};

export const prepareGroupIds = async ({ models, groupId, user }) => {
  if (groupId && groupId !== '0') {
    const access = await models.Group.checkGroupBelongsToUser(user.id, groupId);
    if (!access && !isAdminUser(user)) {
      throw new ValidationError(
        'Access violation.'
      );
    }
    return [groupId];
  }
  // select all user's groups
  const allGroups = await models.Group.findByOwnerId(user.id);
  return allGroups.map(group => group.id);
};

export const downloadFile = async url => {
  try {
    const axiosInstance = axios.create({
      responseType: 'arraybuffer',
      timeout: 30000,
      transformRequest: [(data, headers) => {
        headers.post['User-Agent'] = 'faucet.js';
        return data;
      }],
    });
    const response = await axiosInstance.get(url);
    if (!response.data) return false;
    return response.data;
  } catch (error) {
    console.log(`[downloadFile error] ${url}`);
  }
};

export const toCursorHash = string => Buffer.from(string).toString('base64');

export const fromCursorHash = string => Buffer.from(string, 'base64').toString('ascii');

export const create24HourIntervals = () => {
  const res = [];
  for (let i = 23; i >= 0; i--) {
    const hour = (moment().utc().subtract(i, 'hours')).format('HH');
    res.unshift({ hour, count: 0 });
  }
  return res;
};

export const encrypt = (text, key) => {
  const iv = crypto.randomBytes(16);
  const cipher = crypto.createCipheriv(common.AES_CIPHER, Buffer.from(key), iv);
  let encrypted = cipher.update(text);
  encrypted = Buffer.concat([encrypted, cipher.final()]);
  return `${iv.toString('hex')}${encrypted.toString('hex')}`;
};

export const decrypt = (data, key) => {
  const iv = Buffer.from(data.substr(0, 32), 'hex');
  const encryptedText = Buffer.from(data.substr(32), 'hex');
  const decipher = crypto.createDecipheriv(common.AES_CIPHER, Buffer.from(key), iv);
  let decrypted = decipher.update(encryptedText);
  decrypted = Buffer.concat([decrypted, decipher.final()]);
  return decrypted.toString();
};

export const hmac256FromSeed = async (seed, data) => {
  // const keyHMAC = await crypto.subtle
  //   .importKey('raw', seed, { name: 'HMAC', hash: { name: 'SHA-256' } }, false, ['sign']);
  // return {
  //   hash: await crypto.subtle.sign('HMAC', keyHMAC, data),
  //   seed,
  // };
};

export const RSAEncrypt = (key, data) => {
  return window.crypto.subtle.encrypt(common.KEY_ROLES.ENC.alg, key, data);
};

export const RSAVerify = (key, signature, data) => {
  return crypto.subtle.verify(common.KEY_ROLES.SIGN.alg, key, signature, data);
};
