import 'dotenv/config';
import cors from 'cors';
import morgan from 'morgan';
import http from 'http';
// import https from 'http';
// import fs from 'fs';
import jwt from 'jsonwebtoken';
import DataLoader from 'dataloader';
import express from 'express';
import {
  ApolloServer,
  AuthenticationError,
} from 'apollo-server-express';
import compression from 'compression';
import { createEnfaceApi } from '@enface/sdk';
import multer from 'multer';
// import * as WebSocket from 'ws';
// import { EnfaceAuth } from 'enface-auth-node';
import { EnfaceAuth } from './auth';
import schema from './schema';
import resolvers from './resolvers';
import models, { sequelize } from './models';
import loaders from './loaders';
import * as globals from './common/globals';
import { createDatabaseMock, createAdmin } from './db_mock';
// import { EnfaceAuth } from './common/authbio';
import * as utils from './common/utils';
import * as constants from './common/constants';
import * as bc from './blockchain';
import { airdropCrypto, promoCurrencies } from './blockchain/faucet';

const upload = multer();
const app = express();
app.use(cors());
app.use(morgan('dev'));
app.use(compression());
app.use(express.json());
app.use(express.urlencoded({ extended: true }));

const getMe = async req => {
  const token = req.headers['x-token'];
  if (token) {
    try {
      return await jwt.verify(token, process.env.SECRET);
    } catch (e) {
      throw new AuthenticationError(
        'Your session expired. Sign in again.'
      );
    }
  }
};

export const server = new ApolloServer({
  // cors: false,
  // cors: {
  //   origin: '*',
  //   methods: 'GET,HEAD,PUT,PATCH,POST,DELETE',
  //   preflightContinue: false,
  //   optionsSuccessStatus: 204,
  //   credentials: true,
  // },
  // introspection: true,
  // playground: true,
  introspection: globals.DEBUG,
  playground: globals.DEBUG,
  debug: globals.DEBUG,
  typeDefs: schema,
  resolvers,
  uploads: {
    maxFileSize: 5000000, // 5 MB
    maxFiles: 50,
  },
  formatError: error => {
    const message = error.message
      .replace('SequelizeValidationError: ', '')
      .replace('Validation error: ', '');
    return {
      ...error,
      message,
    };
  },
  context: async (props) => {
    const { req, connection } = props;
    if (connection) {
      return {
        models,
        loaders: {
          user: new DataLoader(keys => loaders.user.batchUsers(keys, models)),
          group: new DataLoader(keys => loaders.group.batchGroups(keys, models)),
          apiKey: new DataLoader(keys => loaders.apiKey.batchApiKeys(keys, models)),
          person: new DataLoader(keys => loaders.person.batchPersons(keys, models)),
        },
        me: connection.context.userFroWS,
      };
    }
    if (req) {
      const me = await getMe(req);
      return {
        models,
        me,
        secret: process.env.SECRET,
        loaders: {
          user: new DataLoader(keys => loaders.user.batchUsers(keys, models)),
          group: new DataLoader(keys => loaders.group.batchGroups(keys, models)),
          apiKey: new DataLoader(keys => loaders.apiKey.batchApiKeys(keys, models)),
          person: new DataLoader(keys => loaders.person.batchPersons(keys, models)),
        },
      };
    }
  },
  subscriptions: {
    onConnect: async (connectionParams) => {
      console.log('+++ subscriptions CONNECTED', connectionParams);
      if (connectionParams.authToken) {
        try {
          const authResult = await jwt.verify(connectionParams.authToken, process.env.SECRET);
          return {
            userFroWS: await models.User.findByLogin(authResult.email),
          };
        } catch (e) {
          throw new AuthenticationError(
            'Wrong or expired auth token!'
          );
        }
      } else if (connectionParams.apiKey) {
        const apiKey = await models.ApiKey.getUserId(connectionParams.apiKey);
        if (!apiKey) {
          throw new AuthenticationError(
            `Access violation [Wrong key ${connectionParams.apiKey}]`
          );
        }
        // console.log('!!!ApiKey.getUserId KEY', apiKey);
        // console.log('!!!user', await models.User.findById(apiKey.userId));
        return {
          userFroWS: await models.User.findById(apiKey.userId),
        };
      }
      throw new Error('Missing auth token!');
    },
    onDisconnect: async () => {
      console.log('--- subscriptions DISCONNECTED');
    },
    keepAlive: 20000,
  },
});

server.applyMiddleware({
  app,
  path: '/graphql',
  // cors: {
  //   origin: '*',
  //   methods: ['OPTIONS', 'POST'],
  //   allowedHeaders: 'DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range',
  //   // preflightContinue: false,
  //   // optionsSuccessStatus: 204,
  //   credentials: true,
  // },
  // cors: false,
});

export const httpServer = http.createServer(app);
// export const httpServer = https.createServer({
//   key: fs.readFileSync(`${__dirname}/ssl/selfsigned.key`),
//   cert: fs.readFileSync(`${__dirname}/ssl/selfsigned.crt`),
// }, app);
server.installSubscriptionHandlers(httpServer);

const port = process.env.PORT || 8000;

new EnfaceAuth({
  httpServer: app,
  // port: 45050,
  projectId: process.env.AUTH_PRODUCT_ID,
  secretCode: process.env.BIO_AUTH_SECRET,
  callbackUrl: 'https://enface-api-server.herokuapp.com',
  // callbackUrl: 'http://localhost:9000',
  // callbackUrl: 'ws://localhost:45050',
  async onCheckCurrentStatus(userId) {
    // const token = jwt.verify(userData, process.env.SECRET);
    if (userId === 'enfaceId') return true;
    console.log('\n[CALLBACK onCheckCurrentStatus]', userId);
    // const userId = token.id;
    const bioUser = await models.AuthBioLink.findByUserId(userId);
    // console.log('Biometric status', bioUser ? bioUser.isActive : false);
    return !!bioUser;
  },
  onUserValidate(userData) {
    console.log('\n[CALLBACK onUserValidate] userData', userData);
    if (userData === 'enfaceId') return 'enfaceId';
    const token = jwt.verify(userData, process.env.SECRET);
    // console.log('[CALLBACK onUserValidate] token', token);
    return token.id;
  },
  async onUserPublicKey(userId) {
    const bioUser = await models.AuthBioLink.findByUserId(userId);
    return bioUser ? bioUser.userPublicKey : null;
  },
  async onActivate(userId, bioId, userPublicKey) {
    console.log('\n[CALLBACK onActivate] userId, bioId', userId, bioId);
    const bioUser = await models.AuthBioLink.findByUserId(userId);
    if (bioUser) {
      await bioUser.destroy({ userId });
      return false;
    }
    await models.AuthBioLink.create({
      userId,
      bioId,
      userPublicKey: userPublicKey || '',
    });
    return true;
  },
  async onUserTokenByBioId(bioId) {
    console.log('\n[CALLBACK onUserTokenByBioId] bioId', bioId);
    const bioUser = await models.AuthBioLink.findByBioId(bioId);
    // console.log('   >>> bioUser userId', !!bioUser);
    if (!bioUser) return false;
    const user = await models.User.findById(bioUser.userId);
    if (!user || !user.isActive) return false;
    return utils.createToken(user, process.env.SECRET, constants.SESSION_JWK_TIMEOUT);
  },
});

sequelize.sync({ force: false }).then(async () => {
  httpServer.listen({ port }, () => {
    console.log(`Apollo Server on http://localhost:${port}/graphql`);
  });
});

// app.get('/persons', async (req, res) => {
//   console.error(req.headers);
//   try {
//     req.headers['x-token'] = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6MSwiZW1haWwiOiJhZG1pbkBlbmZhY2UuaW8iLCJ1c2VybmFtZSI6IkFkbWluIiwicm9sZSI6MTAsImlhdCI6MTU2MDUwODcwNywiZXhwIjoxNTk2NTA4NzA3fQ.MCFM3kveD2MO_voGZVNXCXpElkRQSoH8R55ZKse0NXU';
//     //
//     const authResult = await jwt.verify(req.headers['x-token'], process.env.SECRET);
//     console.error(authResult);
//     const user = await models.User.findByLogin(authResult.email);
//     console.error(user);
//     if (!isAdminUser(user)) {
//       res.send('Wrong or expired auth token!');
//       return;
//     }
//     res.send('PERSONS REQUEST CONTINUE');
//   } catch (e) {
//     res.send('Wrong or expired auth token!');
//   }
// });

// const subscriptionsServer = new SubscriptionServer(
//   {
//     execute,
//     subscribe,
//     schema,
//     onConnect: (connectionParams, webSocket) => {
//       console.log(' !!!!!!!!!+++++++ ON CONNECT!!!!');
//       // if (connectionParams.authToken) {
//       //   return validateToken(connectionParams.authToken)
//       //     .then(findUser(connectionParams.authToken))
//       //     .then((user) => {
//       //       return {
//       //         currentUser: user,
//       //       };
//       //     });
//       // }
//       // throw new Error('Missing auth token!');
//     },
//   },
//   // {
//   //   server: websocketServer,
//   // }
// );

app.route('/rest/*')
  .all((req, res, next) => {
    next();
  })
  .get((req, res) => {
    res.json({});
  })
  .post(upload.array('image', 3), async (req, res) => {
    try {
      if (req.path === '/rest/bcAuthSecret') {
        return res.send(JSON.stringify(await bc.getAuthSecret(req.body.alias)));
      }
      console.log('REST REQUEST:', req.path);
      console.log('POST req.body.apiKey', req.body.apiKey);
      console.log('POST req.files', req.files);
      const enfaceApi = createEnfaceApi({
        apiKey: req.body.apiKey,
      });
      let result;
      switch (req.path) {
        case '/rest/recognize':
          result = await enfaceApi.recognize({
            images: [req.files[0].buffer],
          });
          break;
        case '/rest/liveness':
          result = await enfaceApi.liveness({
            images: [
              req.files[0].buffer,
              req.files[1].buffer,
              req.files[2].buffer,
            ],
          });
          break;
        case '/rest/recognizeLiveness':
          result = await enfaceApi.recognizeLiveness({
            images: [
              req.files[0].buffer,
              req.files[1].buffer,
              req.files[2].buffer,
            ],
          });
          break;
        default:
          break;
      }
      res.send(JSON.stringify(result));
    } catch (error) {
      res.send(JSON.stringify({
        error: `Request error ${error}`,
      }));
    }
  });

app.route('/faucet/*')
  .all((req, res, next) => {
    next();
  })
  .get((req, res) => {
    res.json({});
  })
  .post(upload.none(), async (req, res) => {
    let postData = '';
    req.on('data', chunk => {
      postData += chunk;
    });
    req.on('end', async () => {
      try {
        postData = JSON.parse(postData);
        if (req.path === '/faucet/airdrop') {
          return res.send(JSON.stringify(await airdropCrypto(postData)));
        }
        if (req.path === '/faucet/promo') {
          return res.send(JSON.stringify(await promoCurrencies(postData, req.connection.remoteAddress)));
        }
      } catch (error) {
        res.send(JSON.stringify({
          error: `Request error ${error}`,
        }));
      }
    });
  });
// createAdmin();
