import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter, ForbiddenError } from 'apollo-server';
import Sequelize from 'sequelize';
import uuid4 from 'uuid';
import { isAdmin, isAdminUser } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as constants from '../common/constants';
import { serviceById } from '../services/tools';

const { Op } = Sequelize;

const silenceAllOtherGroupKeys = async (model, mainKey, serviceId) => {
  const sameServiceKeys = await model.findAll({
    where: {
      userId: mainKey.userId,
      serviceId,
      groupId: mainKey.groupId,
      state: {
        [Op.notIn]: [
          constants.KEY_STATE_DEACTIVATED,
          constants.KEY_STATE_SILENT,
        ],
      },
    },
  })
    .filter(item => item.id !== mainKey.id);
  await Promise.all(sameServiceKeys.map(async apiKey => {
    await apiKey.update({
      state: constants.KEY_STATE_SILENT,
    });
    pubsub.publish(EVENTS.API_KEY.UPDATED, {
      apiKeyUpdated: { apiKey },
      userId: apiKey.userId,
    });
  }));
};

export default {
  Query: {
    apiKeys: combineResolvers(isAdmin,
      async (parent, args, { models }) => {
        return models.ApiKey.findAll({
          order: [
            ['id', 'DESC'],
          ],
        });
      }
    ),
    apiKey: async (parent, { uuid }, { models }) => {
      return models.ApiKey.findByUUID(uuid);
    },
    myApiKeys: async (parent, args, { models, me }) => {
      if (!me) {
        throw new ForbiddenError(
          'Access denied'
        );
      }
      return models.ApiKey.findByUserIdActive(me.id);
    },
  },

  Mutation: {

    addApiKey: combineResolvers(
      isAdmin,
      async (
        parent,
        {
          expired, serviceId, actionsLeft,
        },
        { models, me }
      ) => {
        const myMainGroup =  await models.Group.findByOwnerIdFirst(me.id);

        const apiKey = await models.ApiKey.create({
          state: constants.KEY_STATE_SILENT,
          type: constants.KEY_TYPE_CUSTOM,
          expired,
          userId: me.id,
          groupId: myMainGroup.id,
          serviceId,
          actionsLeft,
          uuid: uuid4(),
        });
        pubsub.publish(EVENTS.API_KEY.CREATED, {
          apiKeyCreated: { apiKey },
        });
        return apiKey;
      }
    ),

    updateApiKey: combineResolvers(
      isAdmin,
      async (
        parent,
        {
          id, groupId, expired, state, userId, serviceId, actionsLeft,
        },
        { models }
      ) => {
        const apiKey = await models.ApiKey.findById(id);
        if (!apiKey) {
          throw new ValidationError(
            'Key not found'
          );
        }
        const prevOwner = apiKey.userId;
        const isActivated = apiKey.state === constants.KEY_STATE_DEACTIVATED
          && state !== constants.KEY_STATE_DEACTIVATED;
        const isDeActivated = apiKey.state !== constants.KEY_STATE_DEACTIVATED
          && state === constants.KEY_STATE_DEACTIVATED;
        await apiKey.update({
          groupId,
          expired,
          state: state === constants.KEY_STATE_DEACTIVATED
            ? constants.KEY_STATE_DEACTIVATED
            : constants.KEY_STATE_ACTIVE,
          userId,
          serviceId,
          actionsLeft,
        });
        if (isActivated) {
          pubsub.publish(EVENTS.API_KEY.ADDED, {
            apiKeyAdded: { apiKey },
            userId,
          });
        } else if (isDeActivated) {
          pubsub.publish(EVENTS.API_KEY.REMOVED, {
            apiKeyRemoved: { apiKey },
            userId: prevOwner,
          });
        } else if (
          prevOwner === parseInt(userId, 10)
        ) {
          pubsub.publish(EVENTS.API_KEY.UPDATED, {
            apiKeyUpdated: { apiKey },
            userId,
          });
        } else {
          pubsub.publish(EVENTS.API_KEY.ADDED, {
            apiKeyAdded: { apiKey },
            userId,
          });
          pubsub.publish(EVENTS.API_KEY.REMOVED, {
            apiKeyRemoved: { apiKey },
            userId: prevOwner,
          });
        }
        silenceAllOtherGroupKeys(models.ApiKey, apiKey, serviceId);
        return apiKey;
      }
    ),

    myUpdateApiKey: combineResolvers(
      async (
        parent,
        {
          id, groupId,
        },
        { models, me }
      ) => {
        const apiKey = await models.ApiKey.findById(id);
        if (!apiKey) {
          throw new ValidationError(
            'Key not found'
          );
        }
        if (apiKey.userId !== me.id) {
          throw new ForbiddenError(
            'Access denied'
          );
        }
        if (apiKey.state === constants.KEY_STATE_DEACTIVATED) {
          throw new ForbiddenError(
            'Access denied'
          );
        }
        await apiKey.update({
          groupId,
          state: constants.KEY_STATE_ACTIVE,
        });
        silenceAllOtherGroupKeys(models.ApiKey, apiKey, apiKey.serviceId);
        return apiKey;
      }
    ),
  },

  ApiKey: {
    service: async apiKey => {
      return serviceById(parseInt(apiKey.serviceId, 10));
    },
    user: async (apiKey, args, { loaders }) => {
      return loaders.user.load(parseInt(apiKey.userId, 10));
    },
    group: async (apiKey, args, { loaders }) => {
      return loaders.group.load(parseInt(apiKey.groupId, 10));
    },
  },

  Subscription: {
    apiKeyCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.API_KEY.CREATED),
        (payload, variables, context) => {
          return isAdminUser(context.me);
        }
      ),
    },
    apiKeyAdded: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.API_KEY.ADDED),
        (payload, variables, context) => {
          return context.me.id === parseInt(payload.userId, 10);
        }
      ),
    },
    apiKeyUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.API_KEY.UPDATED),
        (payload, variables, context) => {
          return context.me.id === parseInt(payload.userId, 10);
        }
      ),
    },
    apiKeyRemoved: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.API_KEY.REMOVED),
        (payload, variables, context) => {
          return context.me.id === payload.userId;
        }
      ),
    },
  },
};
