import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter, ForbiddenError } from 'apollo-server';
import moment from 'moment';
import uuid4 from 'uuid';
import { ethers } from 'ethers';
import pubsub, { EVENTS } from '../subscription';
import { isAdmin, isAuthenticated, isAdminUser } from './authorization';
import * as constants from '../common/constants';
import * as utils from '../common/utils';
import * as services from '../services/constants';
import * as processes from '../services/processing';
import { processRecognition } from './serviceFaceAction';
import * as blockchain from '../blockchain';

export default {
  Query: {
    authProject: combineResolvers(
      isAdmin,
      async (parent, { id }, { models }) => {
        return models.AuthProject.findById(id);
      }
    ),
    authProjects: combineResolvers(
      isAuthenticated,
      async (parent, args, { models, me }) => {
        return models.AuthProject.findByOwnerId(me.id);
      }
    ),
  },

  Mutation: {

    createAuthProject: combineResolvers(
      isAuthenticated,
      async (
        parent, { name }, { models, me }
      ) => {
        // console.log(' +++ CREATE ITEM');
        let secret = await utils.getRandomBytes(constants.AUTH_SECRET_LENGTH);
        // console.log(' +++ SECRET1', secret);
        secret = Buffer.from(secret).toString('base64');
        // console.log(' +++ SECRET2', secret);
        const item = await models.AuthProject.create({
          userId: me.id,
          isActive: true,
          name: name || moment().format('MMM, DD YYYY HH:mm:ss'),
          secretKey: secret,
          uuid: uuid4(),
        });
        pubsub.publish(EVENTS.AUTH_PROJECT.CREATED, {
          authProjectCreated: { item },
          userId: me.id,
        });
        // console.log(' +++ RETURNING ITEM', item);
        return item;
      }
    ),

    updateAuthProject: combineResolvers(
      isAuthenticated,
      async (
        parent, {
          id, isActive, name, apiKeyId,
        }, { models, me }
      ) => {
        console.log('[updateAuthProject] isActive', isActive, 'name', name, 'apiKeyId', apiKeyId);
        const item = await models.AuthProject.findById(id);
        if (!item) {
          throw new ValidationError(
            'No auth project found with this id.'
          );
        }
        if (item.userId !== me.id && !isAdminUser(me)) {
          throw new ForbiddenError(
            'Access violation.'
          );
        }
        console.log('isActive', isActive, 'name', name, 'apiKeyId', apiKeyId);
        if (apiKeyId) {
          const apiKey = await models.ApiKey.findById(apiKeyId);
          if (
            !apiKey
            || !models.ApiKey.isActive(apiKey)
            || (apiKey.userId !== me.id && !isAdminUser(me))
          ) {
            throw new ForbiddenError(
              'Wrong api key id.'
            );
          }
        }
        return item.update({
          isActive,
          name,
          apiKeyId,
        });
      }
    ),

    authentication: combineResolvers(
      async (root, {
        files, project, token, fields,
      }, { models, me }) => {
        utils.checkFiles(files, services.AUTH_IMAGES_COUNT);
        utils.logAction(`AUTHENTICATION\n\nproject: ${project}\ntoken: ${token}, fields: ${fields}`);
        const result = { result: false };

        if (!utils.isUuid(project)) {
          result.message = 'Bad project uuid';
          return result;
        }

        const projectDb = await models.AuthProject.findByUuid(project);
        if (!projectDb) {
          result.message = 'Project not found';
          return result;
        }
        console.log('+++ project found', projectDb.id, 'project secret', projectDb.secretKey);
        let decrypted;
        try {
          // token = '733867fa458b005546676831e5b27e145014779609ad688d40d55cec7ee6acbfc341d001f60b3265a8e40fcde61f7a31c1b2ccdc3beb3ae0a4d32aeccc45b0c775471c538f0cdcfa00a4d2fe308c3c4e'; //auth
          // token = '4510e5e6f15a870f80223cb2e17cdd662fb492273d1df704ef77e1088f09fbf7cd6d28bdea7d5ecf73412a8b9d4dff2d41d477c109fa8c16e373d02f639fbf04e64a7dc315ca740819813d85085f975d61e4e5915222ce398ddec35ccf1ce360'; //enable
          decrypted = utils.decrypt(token, Buffer.from(projectDb.secretKey, 'base64'));
        } catch (error) {
          result.message = 'Bad token encryption';
          return result;
        }
        decrypted = decrypted.split('|');
        console.log('+++ decrypted token', decrypted);

        let apiKeyCheck;
        try {
          apiKeyCheck = await utils.checkApiKeyById(projectDb.apiKeyId, models, me);
        } catch (error) {
          result.message = `API key error: ${error.message}`;
          return result;
        }
        const { dbApiKey, user } = apiKeyCheck;
        console.log('+++ API key checked successfully', dbApiKey.uuid);
        utils.updateApiKey(dbApiKey);
        // const sessionId = uuid4();

        let webSocket;
        try {
          webSocket = await processes.checkBackendReady({
            sessionId: decrypted[0],
            url: decrypted[1],
          });

          const images = await utils.getFiles(files, false);
          await utils.checkFiles(images, services.AUTH_IMAGES_COUNT);
          console.log('+++ Project backend ready for biometric data, images:', images);

          if (decrypted[2] === constants.SOCKET_COMMAND_ENABLE) {
            utils.logAction('ENABLING BIOMETRIC AUTHENTICATION');
            let person = null;
            let personUuid = '';

            /* CHECK IDENTIX KNOWS THIS PERSON (get uuid) */
            const [processingResult] = await Promise.all([
              processes.processFaceFind({
                url: services.classify.recognition['1'].url,
                images: [images[0]],
              }),
            ]);

            console.log(' +++ IDENTIX processFace, RESULT', processingResult);
            personUuid = processingResult.uuid;
            if (!personUuid) {
              /* ADD PERSON TO IDENTIX (get uuid) */
              console.log(' --- IDENTIX NO PERSON IN DB', processingResult);
              person = await processes.addPersonIdentix({
                models,
                images: [images[0]],
                userId: dbApiKey.userId,
                groupId: dbApiKey.groupId,
                sourceId: dbApiKey.id,
              });
              personUuid = person.uuid;
              console.log(' >>> IDENTIX PERSON ADDED TO DB', personUuid);
            }

            person = await processes.addPerson({
              models,
              images: [images[0]],
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
              sourceId: dbApiKey.id,
              createDistance: services.classify.recognition['2'].scoring[0].success_strong,
              newPersonUuid: personUuid,
              url: services.classify.recognition['2'].url,
            });

            // publish action to subscribers
            pubsub.publish(EVENTS.PERSON.PERSON_DATA, {
              personData: person,
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            });

            utils.logAction('PERSON SUCCESSFULLY CREATED AND SAVED');
            const postData = {
              _: constants.SOCKET_COMMAND_BIO,
              sessionId: decrypted[0],
              bioId: personUuid,
            };
            const response = webSocket
              ? await webSocket.send(JSON.stringify(postData))
              : await processes.sendAxios(`${decrypted[1]}`, postData);
            utils.logAction(`BIOMETRIC ENABLE BACKEND RESULT ${JSON.stringify(response)}`);
            return {
              result: response._ === 'status',
              action: 'enable',
              message: response._ === 'status'
                ? response.isActive
                  ? 'activated'
                  : 'deactivated'
                : '',
            };
          } if (decrypted[2] === constants.SOCKET_COMMAND_AUTH) {
            utils.logAction('START BIOMETRIC AUTHENTICATION');

            const sessionId = uuid4();
            const version = '2';

            // add new action to db
            const action = await models.FaceAction.create({
              apiKeyId: dbApiKey.id,
              groupId: dbApiKey.groupId,
              userId: user.id,
              sessionId,
              version,
            });

            // publish action to subscribers
            pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
              faceCreated: { action },
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            });

            const resultRecognition = await processRecognition({
              url: services.classify.recognition[version].url,
              images: [images[0]],
              models,
              sessionId,
              saveThumb: true,
              apiKey: dbApiKey,
              action,
              onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
                faceUpdated: { action },
                userId: dbApiKey.userId,
                groupId: dbApiKey.groupId,
              }),
            });
            utils.logAction(`BIOMETRIC AUTHENTICATION RESULT ${resultRecognition.personUuid}`);
            const postData = {
              _: constants.SOCKET_COMMAND_BIO_AUTH,
              sessionId: decrypted[0],
              bioId: resultRecognition.personUuid || '',
            };
            const response = webSocket
              ? await webSocket.send(JSON.stringify(postData))
              : await processes.sendAxios(`${decrypted[1]}`, postData);
            utils.logAction(`BIOMETRIC AUTHORIZATION BACKEND RESULT ${JSON.stringify(response)}`);
            return {
              result: response._ === 'bioauth',
              action: 'auth',
              message: response._ === 'bioauth'
                ? response.result
                  ? 'successful'
                  : 'denied'
                : '',
            };
          }
        } catch (error) {
          utils.logAction(`BIOMETRIC AUTHORIZATION ERROR ${JSON.stringify(error)}`);
          result.message = `Operation failed: ${error.message}`;
          return result;
        } finally {
          webSocket && webSocket.terminate();
        }
      }
    ),

    authBlockchain: combineResolvers(
      async (root, {
        /* files, */alias, security, project, token, fields,
      }, { models, me }) => {
        fields = JSON.parse(fields);
        // utils.checkFiles(files, services.AUTH_BC_IMAGES_COUNT);
        utils.logAction(`AUTHENTICATION\n\nproject: ${project}\ntoken: ${token}, fields: ${fields}, alias: ${alias}, security: ${security}`);
        const result = { result: false };

        if (!utils.isUuid(project)) {
          result.message = 'Bad project uuid';
          return result;
        }

        const projectDb = await models.AuthProject.findByUuid(project);
        if (!projectDb) {
          result.message = 'Project not found';
          return result;
        }
        console.log('+++ project found', projectDb.id, 'project secret', projectDb.secretKey);
        let decrypted;
        try {
          // token = '733867fa458b005546676831e5b27e145014779609ad688d40d55cec7ee6acbfc341d001f60b3265a8e40fcde61f7a31c1b2ccdc3beb3ae0a4d32aeccc45b0c775471c538f0cdcfa00a4d2fe308c3c4e'; //auth
          // token = '4510e5e6f15a870f80223cb2e17cdd662fb492273d1df704ef77e1088f09fbf7cd6d28bdea7d5ecf73412a8b9d4dff2d41d477c109fa8c16e373d02f639fbf04e64a7dc315ca740819813d85085f975d61e4e5915222ce398ddec35ccf1ce360'; //enable
          decrypted = utils.decrypt(token, Buffer.from(projectDb.secretKey, 'base64'));
        } catch (error) {
          result.message = 'Bad token encryption';
          return result;
        }
        decrypted = decrypted.split('|');
        console.log('+++ decrypted token', decrypted);

        let apiKeyCheck;
        try {
          apiKeyCheck = await utils.checkApiKeyById(projectDb.apiKeyId, models, me);
        } catch (error) {
          result.message = `API key error: ${error.message}`;
          return result;
        }
        const { dbApiKey, user } = apiKeyCheck;
        console.log('+++ API key checked successfully', dbApiKey.uuid);
        utils.updateApiKey(dbApiKey);
        // const sessionId = uuid4();

        // check alias ownership

        if (!(await blockchain.checkChallenge(alias, security))) {
          console.log(' --------- SECURITY ERROR ----------------');
          result.message = 'Access violation';
          return result;
        }
        console.log(' ++++++++++ BLOCKCHAIN OWNERSHIP CONFIRMED ++++++++++');

        let webSocket;
        try {
          webSocket = await processes.checkBackendReady({
            sessionId: decrypted[0],
            url: decrypted[1],
          });

          // const images = await utils.getFiles(files, false);
          // await utils.checkFiles(images, services.AUTH_BC_IMAGES_COUNT);
          // console.log('+++ Project backend ready for biometric data, images:', images);

          utils.logAction('START BIOMETRIC AUTHENTICATION');

          // const sessionId = uuid4();
          // const version = '2';
          //
          // // add new action to db
          // const action = await models.FaceAction.create({
          //   apiKeyId: dbApiKey.id,
          //   groupId: dbApiKey.groupId,
          //   userId: user.id,
          //   sessionId,
          //   version,
          // });
          //
          // // publish action to subscribers
          // pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
          //   faceCreated: { action },
          //   userId: dbApiKey.userId,
          //   groupId: dbApiKey.groupId,
          // });
          //
          // const resultRecognition = await processRecognition({
          //   url: services.classify.recognition[version].url,
          //   images: [images[0]],
          //   models,
          //   sessionId,
          //   saveThumb: true,
          //   apiKey: dbApiKey,
          //   action,
          //   onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
          //     faceUpdated: { action },
          //     userId: dbApiKey.userId,
          //     groupId: dbApiKey.groupId,
          //   }),
          // });
          // if (!resultRecognition.person) {
          //   result.message = 'Unknown person';
          //   return result;
          // }
          // console.error(` ++++++++ PERSON bcId ${resultRecognition.person.bcId} ++++++++ `);
          // if (resultRecognition.person.bcId !== alias) {
          //   result.message = 'This is not you';
          //   return result;
          // }
          console.error(' ++++++++ PERSON BIOMETRIC SUCCCESS ++++++++ ');

          // get user ethereum wallet address
          const userWallet = await blockchain.getUserWallet(blockchain.wallet.contract, alias);
          console.error(` ++++++++ USER WALLET ${userWallet} ++++++++ `);

          const [usd, eur, rub] = await Promise.all([
            blockchain.wallet.usd.balanceOf(userWallet),
            blockchain.wallet.eur.balanceOf(userWallet),
            blockchain.wallet.rub.balanceOf(userWallet),
          ]);

          // utils.logAction(`BIOMETRIC AUTHENTICATION RESULT ${resultRecognition.personUuid}`);
          const postData = {
            _: constants.SOCKET_COMMAND_BLOCKCHAIN_AUTH,
            sessionId: decrypted[0],
            alias,
            fields: {
              ...fields,
              usd: ethers.utils.formatEther(usd).toString(),
              eur: ethers.utils.formatEther(eur).toString(),
              rub: ethers.utils.formatEther(rub).toString(),
            },
          };
          const response = webSocket
            ? await webSocket.send(JSON.stringify(postData))
            : await processes.sendAxios(`${decrypted[1]}`, postData);
          utils.logAction(`BIOMETRIC AUTHORIZATION BACKEND RESULT ${JSON.stringify(response)}`);
          return {
            result: response._ === 'bioauth',
            action: 'auth',
            message: response._ === 'bioauth'
              ? response.result
                ? 'successful'
                : 'denied'
              : '',
          };
        } catch (error) {
          utils.logAction(`BIOMETRIC AUTHORIZATION ERROR ${JSON.stringify(error)}`);
          result.message = `Operation failed: ${error.message}`;
          return result;
        } finally {
          webSocket && webSocket.terminate();
        }
      }
    ),

    payment: combineResolvers(
      async (root, {
        /* files, */ alias, security, projectId, productId, txId,
      }, { models, me }) => {
        // utils.checkFiles(files, services.PAYMENT_IMAGES_COUNT);
        utils.logAction(`PAYMENT\n\nproject: ${projectId}\nproductId: ${productId}, txId: ${txId}, alias: ${alias}, security: ${security}`);
        const result = { result: false };

        if (!utils.isUuid(projectId)) {
          result.message = 'Bad project uuid';
          return result;
        }

        const projectDb = await models.AuthProject.findByUuid(projectId);
        if (!projectDb) {
          result.message = 'Project not found';
          return result;
        }
        console.log('+++ project found', projectDb.id, 'project secret', projectDb.secretKey);
        // let decrypted;
        // try {
        //   // token = '733867fa458b005546676831e5b27e145014779609ad688d40d55cec7ee6acbfc341d001f60b3265a8e40fcde61f7a31c1b2ccdc3beb3ae0a4d32aeccc45b0c775471c538f0cdcfa00a4d2fe308c3c4e'; //auth
        //   // token = '4510e5e6f15a870f80223cb2e17cdd662fb492273d1df704ef77e1088f09fbf7cd6d28bdea7d5ecf73412a8b9d4dff2d41d477c109fa8c16e373d02f639fbf04e64a7dc315ca740819813d85085f975d61e4e5915222ce398ddec35ccf1ce360'; //enable
        //   decrypted = utils.decrypt(token, Buffer.from(projectDb.secretKey, 'base64'));
        // } catch (error) {
        //   result.message = 'Bad token encryption';
        //   return result;
        // }
        // decrypted = decrypted.split('|');
        // console.log('+++ decrypted token', decrypted);

        let apiKeyCheck;
        try {
          apiKeyCheck = await utils.checkApiKeyById(projectDb.apiKeyId, models, me);
        } catch (error) {
          result.message = `API key error: ${error.message}`;
          return result;
        }
        const { dbApiKey, user } = apiKeyCheck;
        console.log('+++ API key checked successfully', dbApiKey.uuid);
        utils.updateApiKey(dbApiKey);
        // const sessionId = uuid4();

        // check alias ownership

        if (!(await blockchain.checkChallenge(alias, security))) {
          console.log(' --------- SECURITY ERROR ----------------');
          result.message = 'Access violation';
          return result;
        }
        console.log(' ++++++++++ BLOCKCHAIN OWNERSHIP CONFIRMED ++++++++++');

        // let webSocket;
        let product;
        try {
          // webSocket = await processes.checkBackendReady({
          //   sessionId: decrypted[0],
          //   url: decrypted[1],
          // });

          for (const amount in constants.TEST_PAYMENTS) {
            if (constants.TEST_PAYMENTS[amount].id === productId) {
              product = {
                ...constants.TEST_PAYMENTS[amount],
                price: amount,
              };
              break;
            }
          }
          if (!product) {
            console.log(` --------- PRODUCT NOT FOUND ${product} ----------------`);
            result.message = 'Bad product id';
            return result;
          }
          utils.logAction(`PRODUCT FOUND ${product.url}`);

          // const images = await utils.getFiles(files, false);
          // await utils.checkFiles(images, services.AUTH_BC_IMAGES_COUNT);
          // console.log('+++ Project backend ready for biometric data, images:', images);
          //
          // utils.logAction('START BIOMETRIC AUTHENTICATION');
          //
          // const sessionId = uuid4();
          // const version = '2';
          //
          // // add new action to db
          // const action = await models.FaceAction.create({
          //   apiKeyId: dbApiKey.id,
          //   groupId: dbApiKey.groupId,
          //   userId: user.id,
          //   sessionId,
          //   version,
          // });
          //
          // // publish action to subscribers
          // pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
          //   faceCreated: { action },
          //   userId: dbApiKey.userId,
          //   groupId: dbApiKey.groupId,
          // });
          //
          // const resultRecognition = await processRecognition({
          //   url: services.classify.recognition[version].url,
          //   images: [images[0]],
          //   models,
          //   sessionId,
          //   saveThumb: true,
          //   apiKey: dbApiKey,
          //   action,
          //   onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
          //     faceUpdated: { action },
          //     userId: dbApiKey.userId,
          //     groupId: dbApiKey.groupId,
          //   }),
          // });
          // if (!resultRecognition.person) {
          //   result.message = 'Unknown person';
          //   return result;
          // }
          // console.error(` ++++++++ PERSON bcId ${resultRecognition.person.bcId} ++++++++ `);
          // if (resultRecognition.person.bcId !== alias) {
          //   result.message = 'This is not you';
          //   return result;
          // }
          // console.error(' ++++++++ PERSON BIOMETRIC SUCCCESS ++++++++ ');

          // get user ethereum wallet address
          const userWallet = await blockchain.getUserWallet(blockchain.wallet.contract, alias);
          console.error(` ++++++++ USER WALLET ${userWallet} ++++++++ `);

          result.message = 'success';
          return {
            result: true,
            action: 'paymeent',
            message: 'success',
          };
        } catch (error) {
          utils.logAction(`BIOMETRIC AUTHORIZATION ERROR ${JSON.stringify(error)}`);
          result.message = `Operation failed: ${error.message}`;
          return result;
        } finally {
          // webSocket && webSocket.terminate();
        }
      }
    ),

  },

  AuthProject: {
    apiKey: async (item, args, { models }) => {
      // console.log('>>>>>>>>>> item.apiKeyId', item.apiKeyId);
      const apiKey = item.apiKeyId ? await models.ApiKey.findById(item.apiKeyId, 10) : null;
      // console.log('>>>>>>>>>> apiKey', apiKey);
      return apiKey && models.ApiKey.isActive(apiKey) ? apiKey : null;
    },
  },

  Subscription: {
    authProjectCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.AUTH_PROJECT.CREATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId;
        }
      ),
    },
  },
};
