import uuid4 from 'uuid';
import { combineResolvers } from 'graphql-resolvers';
import { withFilter, UserInputError, ValidationError } from 'apollo-server';
import * as utils from '../common/utils';
import { isAdmin, isAdminUser } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as services from '../services/constants';
import * as processes from '../services/processing';
import * as constants from '../common/constants';

export default {
  Query: {
    guestActions: combineResolvers(
      isAdmin,
      async (parent, args, { models }) => {
        return models.GuestAction.findAll();
      }
    ),
  },

  Mutation: {

    signInFace: combineResolvers(
      async (root, { files }, { models, me, secret }) => {
        // check is guest
        if (me) {
          throw new UserInputError(
            'You are already authorized'
          );
        }

        // preparations
        const images = await utils.getFiles(files);
        await utils.checkFiles(images, services.SIGNIN_IMAGES);
        const sessionId = uuid4();

        utils.logAction('Biometric SignIn');

        // add new action to db
        const action = await models.GuestAction.create({
          sessionId,
          action: services.GUEST_ACTION_SIGNIN,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.GUEST.GUEST_ACTION_CREATED, {
          guestActionCreated: { action },
        });

        // await for all processings done
        const [checkResult] = await Promise.all([
          processes.processFaceFind({
            images,
          }),
          utils.processAWS({
            sessionId,
            images,
          }),
        ]);

        console.log(' +++ ALL ACTIONS DONE, RESULT', checkResult);

        if (
          !checkResult.value
          || !checkResult.value
          || checkResult.value < 0
          || !checkResult.idxid
          || checkResult.value !== services.IDENTIX_EXACT
        ) {
          throw new ValidationError(1);
        }

        const person = await models.Person.findByUUID(checkResult.idxid);
        if (!person) {
          utils.logAction('Person not found!');
          throw new ValidationError(2);
        }
        utils.logAction(`Person found! [uuid: ${checkResult.idxid}, userId: ${person.userId}, typeof: ${typeof person.userId}]`);
        const user = await models.User.findById(person.userId);
        if (!user) {
          utils.logAction('No user with this person linked');
          throw new ValidationError(3);
        }

        if (!user.biometricSignIn) {
          throw new ValidationError(4);
        }

        // save thumb to aws
        utils.processAWSThumb({
          imageBuffer: checkResult.thumb,
          sessionId,
          prefix: '',
          onUpdateAction: () => pubsub.publish(EVENTS.GUEST.GUEST_ACTION_UPDATED, {
            guestActionUpdated: { action },
          }),
        });

        return {
          token: await utils.createToken(user, secret, constants.SESSION_JWK_TIMEOUT),
          statsToken: await utils.createStatsToken(user, secret),
        };
      }
    ),

  },

  GuestAction: {
    user: async (action, args, { loaders }) => {
      if (!action.userId) {
        return {
          id: 0,
          username: 'Guest',
        };
      }
      return loaders.user.load(parseInt(action.userId, 10));
    },
    imageUrl: action => {
      if (action.result === null) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_thumb_50x50.jpg`;
    },
    imageOriginalUrl: action => {
      if (action.result === null) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_0.jpg`;
    },
  },

  Subscription: {
    guestActionCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.GUEST.GUEST_ACTION_CREATED),
        (payload, variables, context) => {
          return !!isAdminUser(context.me);
        }
      ),
    },
    guestActionUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.GUEST.GUEST_ACTION_UPDATED),
        (payload, variables, context) => {
          return !!isAdminUser(context.me);
        }
      ),
    },
  },
};
