import Sequelize from 'sequelize';
import { combineResolvers } from 'graphql-resolvers';
import { withFilter, ValidationError } from 'apollo-server';
import { isAuthenticated } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as constants from '../common/constants';

const { Op } = Sequelize;

export default {

  Query: {

    joinRequests: combineResolvers(
      isAuthenticated,
      async (parent, args, { models, me }) => {
        return models.JoinRequest.findAll({
          where: {
            [Op.or]: [{ userId: me.id }, { fromUserId: me.id }],
          },
          order: [
            ['createdAt', 'DESC'],
          ],
        });
      }
    ),
  },

  Mutation: {

    createJoinRequest: combineResolvers(
      isAuthenticated,
      async (
        parent, { email }, { models, me }
      ) => {
        const user = await models.User.findByLogin(email);
        if (!user) {
          throw new ValidationError(
            `User with email "${email}" not found.`
          );
        }
        if (user.id === me.id) {
          throw new ValidationError(
            'You can not join yourself.'
          );
        }
        try {
          const joinRequest = await models.JoinRequest.create({
            fromUserId: me.id,
            userId: user.id,
            state: constants.JOIN_STATE_NEW,
          });
          pubsub.publish(EVENTS.JOIN_REQUEST.JOIN_CREATED, {
            joinRequestCreated: joinRequest,
            fromUserId: me.id,
            userId: user.id,
          });
          return joinRequest;
        } catch (error) {
          throw new ValidationError(
            'Your have already sent the request to this team.'
          );
        }
      }
    ),

    updateJoinRequest: combineResolvers(
      isAuthenticated,
      async (
        parent, { id, state }, { models, me }
      ) => {
        const joinRequest = await models.JoinRequest.findById(id);
        if (!joinRequest) {
          throw new ValidationError(
            'No request found with this id'
          );
        }
        if ( // user revoking or re-submitting his request
          [
            constants.JOIN_STATE_CANCELLED,
            constants.JOIN_STATE_NEW,
          ].includes(state)
        ) {
          if (
            joinRequest.fromUserId !== me.id
            // user tries to resend rejected request
            || joinRequest.state === constants.JOIN_STATE_DECLINED
          ) {
            throw new ValidationError(
              'Access violation'
            );
          }
        } else if ( // host user changing status
          [
            constants.JOIN_STATE_ACCEPTED,
            constants.JOIN_STATE_DECLINED,
          ].includes(state)
        ) {
          if (joinRequest.userId !== me.id) {
            throw new ValidationError(
              'Access violation'
            );
          }
        } else {
          throw new ValidationError(
            'Wrong incoming params'
          );
        }
        try {
          await joinRequest.update({
            state,
          });

          // SUBSCRIPTIONS
          pubsub.publish(EVENTS.JOIN_REQUEST.JOIN_UPDATED, {
            joinRequestUpdated: joinRequest,
            fromUserId: joinRequest.fromUserId,
            userId: joinRequest.userId,
          });
          return joinRequest;
        } catch (error) {
          throw new ValidationError(
            'Internal server error'
          );
        }
      }
    ),
  },

  JoinRequest: {
    toUser: async (joinRequest, args, { loaders }) => {
      return loaders.user.load(joinRequest.userId);
    },
    fromUser: async (joinRequest, args, { loaders }) => {
      return loaders.user.load(joinRequest.fromUserId);
    },
  },

  Subscription: {
    joinRequestCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.JOIN_REQUEST.JOIN_CREATED),
        (payload, variables, context) => {
          return [payload.userId, payload.fromUserId].includes(context.me.id);
        }
      ),
    },
    joinRequestUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.JOIN_REQUEST.JOIN_UPDATED),
        (payload, variables, context) => {
          return [payload.userId, payload.fromUserId].includes(context.me.id);
        }
      ),
    },
  },
};
