import Sequelize from 'sequelize';
import { combineResolvers } from 'graphql-resolvers';
import { withFilter, ValidationError } from 'apollo-server';
import * as constants  from '../common/constants';
import * as utils from '../common/utils';
import pubsub, { EVENTS } from '../subscription';
import * as services from '../services/constants';
import * as processes from '../services/processing';
import { isAdminUser } from './authorization';

const { Op } = Sequelize;

export default {
  Query: {
    persons: combineResolvers(
      async (parent, {
        groupId, token, cursor, limit = 100,
      }, { models, me, secret }) => {
        const user = me || await utils.validateStatsToken(token, secret);
        const cursorOptions = cursor
          ? {
            where: {
              id: {
                [Sequelize.Op.lt]: utils.fromCursorHash(cursor),
              },
            },
          }
          : {};
        cursorOptions.where = { ...cursorOptions.where, groupId: { [Op.not]: null } };
        if (isAdminUser(user) && groupId === undefined) {
          // all persons
        } else {
          const groupIds = await utils.prepareGroupIds({ models, groupId, user });
          cursorOptions.where = { ...cursorOptions.where, groupId: { [Op.in]: groupIds } };
        }
        const result = await models.Person.findAll({
          order: [
            ['createdAt', 'DESC'],
          ],
          limit: limit + 1,
          ...cursorOptions,
        });
        const hasNextPage = result.length > limit;
        const edges = hasNextPage ? result.slice(0, -1) : result;
        return {
          edges,
          pageInfo: {
            hasNextPage,
            endCursor: utils.toCursorHash(edges.length ? edges[edges.length - 1].id.toString() : ''),
          },
        };
      }
    ),
  },

  Mutation: {

    addPerson: combineResolvers(
      async (root, {
        files, apiKey, imageURL, version,
      }, { models, me }) => {
        // preparations
        const { dbApiKey } = await utils.checkApiKey(apiKey, models, me);
        let images = [];
        if (imageURL) {
          const downloaded = await utils.downloadFile(imageURL);
          if (!downloaded) {
            throw new ValidationError('Bad URL for image provided');
          }
          images.push(downloaded);
        } else {
          images = await utils.getFiles(files, false);
          await utils.checkFiles(images, services.PERSON_IMAGES_MIN);
        }
        if (!dbApiKey.groupId) {
          throw new ValidationError('apiKey is not assigned to any group!');
        }
        utils.logAction(`ADD PERSON [version: ${version}] ${images.length} ${images[0].length}`);
        await utils.updateApiKey(dbApiKey);

        if (version === '1') {
          try {
            const person = await processes.addPersonIdentix({
              models,
              images,
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
              sourceId: dbApiKey.id,
            });
            // publish action to subscribers
            pubsub.publish(EVENTS.PERSON.PERSON_DATA, {
              personData: person,
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            });

            return person;
          } catch (error) {
            throw new ValidationError(error);
          }
        } else {
          try {
            let person = null;
            let personUuid = '';

            // await uuid from identix
            const [processingResult] = await Promise.all([
              processes.processFaceFind({
                url: services.classify.recognition['1'].url,
                images,
              }),
            ]);

            console.log(' +++ IDENTIX processFace, RESULT', processingResult);
            personUuid = processingResult.uuid;
            if (!personUuid) {
              console.log(' --- IDENTIX NO PERSON IN DB', processingResult);
              person = await processes.addPersonIdentix({
                models,
                images,
                userId: dbApiKey.userId,
                groupId: dbApiKey.groupId,
                sourceId: dbApiKey.id,
              });
              personUuid = person.uuid;
              console.log(' >>> IDENTIX PERSON ADDED TO DB', personUuid);
            }
            // personUuid = await utils.md5(`${personUuid}${dbApiKey.groupId}`);
            // console.log(' >>> HASHED >>>>', personUuid);
            // personUuid = utils.string2uuid(personUuid);
            // console.log(' >>> HASHED UUID >>>>', personUuid);
            //
            // person = await models.Person.findWithParams({
            //   uuid: processingResult.uuid,
            //   groupId: dbApiKey.groupId,
            //   userId: dbApiKey.userId,
            // });
            // if (!person) {
            //   // throw new ValidationError('Person with this UUID not found!');
            // }

            person = await processes.addPerson({
              models,
              images,
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
              sourceId: dbApiKey.id,
              createDistance: services.classify.recognition['2'].scoring[0].success_strong,
              newPersonUuid: personUuid,
              url: services.classify.recognition['2'].url,
            });

            // publish action to subscribers
            pubsub.publish(EVENTS.PERSON.PERSON_DATA, {
              personData: person,
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            });

            return person;
          } catch (error) {
            throw new ValidationError(error);
          }
        }
      }
    ),

  },

  Person: {
    user: (item, args, { loaders }) => {
      return loaders.user.load(parseInt(item.userId, 10));
    },
    group: (item, args, { loaders }) => {
      return loaders.group.load(parseInt(item.groupId, 10));
    },
    apiKey: (item, args, { loaders }) => {
      return loaders.apiKey.load(parseInt(item.sourceId, 10));
    },
    thumb50x50: item => {
      if (!item.uuid) return '';
      const res = utils.createImagePath(item.uuid);
      return `${constants.AWS_BUCKET_URL}${constants.AWS_PERSONS_PREFIX}${res.path}${res.name}_thumb_50x50.jpg?${utils.getUnixTimeStamp(item.updatedAt)}`;
    },
    thumb200x200: item => {
      if (!item.uuid) return '';
      const res = utils.createImagePath(item.uuid);
      return `${constants.AWS_BUCKET_URL}${constants.AWS_PERSONS_PREFIX}${res.path}${res.name}_thumb_200x200.jpg?${utils.getUnixTimeStamp(item.updatedAt)}`;
    },
    imageURL: item => {
      if (!item.uuid) return '';
      const res = utils.createImagePath(item.uuid);
      return `${constants.AWS_BUCKET_URL}${constants.AWS_PERSONS_PREFIX}${res.path}${res.name}_0.jpg`;
    },
  },

  Subscription: {
    personData: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.PERSON.PERSON_DATA),
        (payload, variables, context) => {
          return  context.me.id === payload.userId
            && (!payload.groupId || payload.groupId === '0' || payload.groupId === parseInt(variables.groupId, 10));
        }
      ),
    },
  },
};
