import uuid4 from 'uuid';
import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter } from 'apollo-server';
import { ethers } from 'ethers';
import * as utils from '../common/utils';
import { isAdminUser } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as constants from '../common/constants';
import * as services from '../services/constants';
import * as processes from '../services/processing';
import * as tools from '../services/tools';
import * as blockchain from '../blockchain';

export const processFace = async ({
  url,
  images,
  models,
  sessionId,
  apiKey,
  action,
  resultPrefix = '',
  onPublish,
}) => {
  const started = new Date();
  let person = null;

  // saving thumbs
  utils.processAWS({
    sessionId,
    images,
    // onUpdateAction: onPublish,
  });

  // await for all processings done
  const [processingResult] = await Promise.all([
    processes.processFaceFind({
      url,
      images,
    }),
  ]);

  console.log(' +++ IDENTIX DONE, RESULT', processingResult);

  if (processingResult.value >= services.IDENTIX_JUNK) {
    const personUuid = processingResult.uuid;
    // let personUuid = await utils.md5(`${processingResult.uuid}${apiKey.groupId}`);
    // personUuid = utils.string2uuid(personUuid);
    person = await models.Person.findWithParams({
      uuid: personUuid,
      groupId: apiKey.groupId,
      userId: apiKey.userId,
    });
    if (!person) processingResult.value = 0;
  }

  // publish action to subscribers
  await action.update({
    [`time${resultPrefix}`]: (new Date() - started) / 1000,
    [`result${resultPrefix}`]: processingResult.value,
    personId: person ? person.id : null,
  });

  // save images to aws
  utils.processAWSThumb({
    imageBuffer: processingResult.thumb,
    sessionId,
    prefix: '',
    onUpdateAction: onPublish,
  });
  return action;
};

export const processRecognition = async ({
  url,
  images,
  models,
  sessionId,
  apiKey,
  action,
  resultPrefix = '',
  onPublish,
}) => {
  const started = new Date();
  let person = null;
  console.log(' +++ ENFACE START');

  // saving thumbs
  utils.processAWS({
    sessionId,
    images,
    // onUpdateAction: onPublish,
  });

  // await for all processings done
  const [processingResult] = await Promise.all([
    processes.processRecognition({
      userId: apiKey.userId,
      // userId: 2,
      groupId: apiKey.groupId,
      sessionId,
      // // createDistance: 0,
      // newPersonUuid: sessionId,
      url,
      images,
    }),
  ]);

  console.log(' +++ ENFACE DONE, RESULT', processingResult);

  // if (processingResult.value >= services.classify.recognition['2'].scoring[0].success_poor) {
  if (
    processingResult.value >= 0
    && processingResult.value < services.classify.recognition['2'].scoring[1].success_poor
  ) {
    // console.log(' +++ should search for person');
    try {
      person = await models.Person.findWithParams({
        uuid: processingResult.uuid,
        groupId: apiKey.groupId,
        userId: apiKey.userId,
      });
      console.log(` +++ person search result ${person ? person.id : 'null'}`);
      action.person = person;
      if (!person) {
        processingResult.value = services.classify.recognition['2'].scoring[2].failed_strong;
      }
    } catch (error) {
      // console.log(' !!! person search result ERROR', error);
      // console.log(' !!! person search result ERROR');
      processingResult.value = services.SERVICE_ERROR_WRONG_RESPONSE;
    }
  }

  // publish action to subscribers
  await action.update({
    [`time${resultPrefix}`]: (new Date() - started) / 1000,
    [`result${resultPrefix}`]: processingResult.value,
    personId: person ? person.id : null,
  });

  // save images to aws
  utils.processAWSThumb({
    imageBuffer: processingResult.thumb,
    sessionId,
    prefix: '',
    onUpdateAction: onPublish,
  });
  console.log('>>>>RECOGNITION DONE', JSON.stringify(action));
  return action;
};

export default {
  Query: {
    faceAction: combineResolvers(
      async (parent, { id, token }, { models, me, secret }) => {
        const user = me || await utils.validateStatsToken(token, secret);
        const action = await models.FaceAction.findById(id);
        if (!action) {
          throw new ValidationError(
            'No records found.'
          );
        }
        if (action.userId !== user.id && !isAdminUser(user)) {
          throw new ValidationError(
            'Access violation.'
          );
        }
        return action;
      }
    ),
    faceActions: combineResolvers(
      async (parent, {
        groupId, token, cursor, limit = 100,
      }, { models, me, secret }) => {
        return processes.getServiceResults({
          user: me || await utils.validateStatsToken(token, secret),
          serviceModel: models.FaceAction,
          groupId,
          cursor,
          limit,
          models,
        });
      }
    ),
  },

  Mutation: {
    goFace: combineResolvers(
      async (root, { files, apiKey, version }, { models, me }) => {
        utils.checkFiles(files, services.FACE_IMAGES);
        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        let sessionId = uuid4();
        version = tools.checkVersion('recognition', version, isAdminUser(user));

        version = '1';

        // add new action to db
        const action1 = await models.FaceAction.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: user.id,
          sessionId,
          version,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
          faceCreated: { action: action1 },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        // await utils.checkFiles(images, services.FACE_IMAGES);

        processFace({
          url: services.classify.recognition[version].url,
          images,
          models,
          sessionId,
          saveThumb: true,
          apiKey: dbApiKey,
          action: action1,
          onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
            faceUpdated: { action: action1 },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        version = '2';
        sessionId = uuid4();

        // add new action to db
        const action2 = await models.FaceAction.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: user.id,
          sessionId,
          version,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
          faceCreated: { action: action2 },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        // await utils.checkFiles(images, services.FACE_IMAGES);
        processRecognition({
          url: services.classify.recognition['2'].url,
          images,
          models,
          sessionId,
          saveThumb: true,
          apiKey: dbApiKey,
          action: action2,
          onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
            faceUpdated: { action: action2 },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        return action2;
      }
    ),

    recognize: combineResolvers(
      async (root, {
        files, apiKey, version, pc, npc, bcId, pa, security,
      }, { models, me }) => {
        utils.checkFiles(files, services.FACE_IMAGES);
        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        version = tools.checkVersion('recognition', version, isAdminUser(user));

        // version = '1';
        // let sessionId = uuid4();
        //
        // // add new action to db
        // const action2 = await models.FaceAction.create({
        //   apiKeyId: dbApiKey.id,
        //   groupId: dbApiKey.groupId,
        //   userId: user.id,
        //   sessionId,
        //   version,
        // });
        //
        // // publish action to subscribers
        // pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
        //   faceCreated: { action: action2 },
        //   userId: dbApiKey.userId,
        //   groupId: dbApiKey.groupId,
        // });
        //
        // const images = await utils.getFiles(files);
        // // await utils.checkFiles(images, services.FACE_IMAGES);
        //
        // // await utils.checkFiles(images, services.FACE_IMAGES);
        // processFace({
        //   url: services.classify.recognition[version].url,
        //   images,
        //   models,
        //   sessionId,
        //   saveThumb: true,
        //   apiKey: dbApiKey,
        //   action: action2,
        //   onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
        //     faceUpdated: { action: action2 },
        //     userId: dbApiKey.userId,
        //     groupId: dbApiKey.groupId,
        //   }),
        // });

        version = '2';
        const sessionId = uuid4();

        // add new action to db
        const action1 = await models.FaceAction.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: user.id,
          sessionId,
          version,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
          faceCreated: { action: action1 },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        // await utils.checkFiles(images, services.FACE_IMAGES);

        const result = await processRecognition({
          url: services.classify.recognition[version].url,
          images,
          models,
          sessionId,
          saveThumb: true,
          apiKey: dbApiKey,
          action: action1,
          onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
            faceUpdated: { action: action1 },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        // application passcode & person creation
        if (pc || npc) {
          utils.logAction(`
            processFace
            blockchainId: ${bcId}
            operation: ${pc}
            newPassCode: ${npc}
            paymentAmount: ${pa}
          `);
          if (pa) { // processing payment
            console.error(` >>> PROCESSING PAYMENT WITH PERSON ${result.person ? result.person.id : 'null'}`);
            let paymentAmount;
            try {
              paymentAmount = pa.toString();
              console.error(' >>> PROCESSING PAYMENT', pa);
              if (!constants.TEST_PAYMENTS[pa]) {
                result.actionCode = constants.PAYMENT_ERRORS['0'];
                return result;
              }
              console.error(' +++ PAYMENT AMOUNT', paymentAmount, 'url', constants.TEST_PAYMENTS[paymentAmount].url);
              if (!result.person) {
                result.actionCode = constants.PAYMENT_ERRORS['1'];
                return result;
              }
              console.error(' +++ PAYMENT REAL PASSCODE', result.person.passCode, 'GOT', pc);
              if (result.person.passCode !== pc) {
                result.actionCode = constants.PAYMENT_ERRORS['2'];
                return result;
              }
            } catch (error) {
              console.error(' ! [ERROR]', error);
              result.actionCode = constants.PAYMENT_ERRORS['4'];
              return result;
            }
            try {
              // get user ethereum wallet address
              const userWallet = await blockchain.getUserWallet(blockchain.wallet.contract, result.person.bcId);
              // const userWallet = '0xeF4FC67ACCe24c1d3FEc31D92717c6410a3D533A';

              console.log(`blockchain user wallet ${userWallet}`);
              // get allowance balance to pay from
              const allowance = await blockchain.wallet.usd.allowance(userWallet, blockchain.wallet.address);

              console.log('allowance', allowance.toString());
              console.log('burning  ', ethers.utils.parseEther(paymentAmount).toString());
              // check allowance balance to pay from
              if (allowance.lt(ethers.utils.parseEther(paymentAmount))) {
                result.actionCode = constants.PAYMENT_ERRORS['3'];
                return result;
              }

              // burn the payment amount from user balance
              const burnResult = await blockchain
                .wallet
                .usd
                .burnFrom(userWallet, ethers.utils.parseEther(paymentAmount));

              console.log('burnResult', burnResult);

              // return success with actionCode
              result.actionCode = 'success';
              // result.actionCode = constants.TEST_PAYMENTS[paymentAmount].url;
              // utils.downloadFile(constants.TEST_PAYMENTS[paymentAmount].url);
              return result;
            } catch (error) {
              console.error(' ! [BLOCKCHAIN ERROR]', error);
              result.actionCode = constants.PAYMENT_ERRORS['5'];
              return result;
            }
          }

          // check bcId ownership

          if (!(await blockchain.checkChallenge(bcId, security))) {
            console.log(' --------- SECURITY ERROR ----------------');
            result.actionCode = constants.PAYMENT_ERRORS['6'];
            return result;
          }
          // security
          const person = await models.Person.findWithParams({
            groupId: dbApiKey.groupId,
            userId: dbApiKey.userId,
            bcId,
          });
          console.error(` !!!!!!!!!!!!!!!!!! PERSON ID BY bcId ${bcId} ${person ? person.id : 'NULL'}`);
          if (
            person
              && (
                (result.person && person.id !== result.person.id)
                || (!result.person)
              )
          ) await person.destroy();
          if (result.person) {
            console.log(`person found ${result.person.uuid}, checking bcId`);
            if (result.person.bcId === bcId) {
              // console.error(` !!!!!!!!!!!!!!!!!! result.person.bcId === bcId ${result.person.bcId} ${bcId}`);
            } else {
              // console.error(` !!!!!!!!!!!!!!!!!! result.person.bcId !== bcId ${result.person.bcId} ${bcId}`);
              await result.person.update({
                bcId,
              });
            }
          }
          console.error(` !!!!!!!!!!!!!!!!!! PERSON ID AFTER ALL ${result.person ? result.person.id : 'NULL'}`);
          utils.logAction(`
            person with this blockchain id
            person: ${result.person ? result.person.id : 'NULL'}
          `);
          result.person = result.person || await processes.addPerson({
            models,
            images,
            bcId,
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
            sourceId: dbApiKey.id,
            createDistance: services.classify.recognition['2'].scoring[0].success_strong,
            newPersonUuid: uuid4(),
            url: services.classify.recognition['2'].url,
          });
          console.log(`person before passcode & bcId update ${result.person.id}`);
          await result.person.update({
            bcId,
            passCode: npc,
          });
        }
        console.log('ALL DONE OK');
        return result;
      }
    ),

    //     goFace: combineResolvers(
    //       async (root, { files, apiKey, version }, { models, me }) => {
    //         console.log('\n\n[goFace]', files, apiKey, version);
    //         utils.checkFiles(files, services.FACE_IMAGES);
    //         // preparations
    //         const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
    //         // console.log('\n\n[goFace] dbApiKey', dbApiKey, 'user', user);
    //         utils.updateApiKey(dbApiKey);
    //         let sessionId = uuid4();
    //         version = tools.checkVersion('recognition', version, isAdminUser(user));
    //
    //         version = '1';
    //         const images = await utils.getFiles(files); // to_del
    //         // console.log('\n\n[goFace] +++ALL READY+++');
    //
    //         // // add new action to db
    //         // const action1 = await models.FaceAction.create({
    //         //   apiKeyId: dbApiKey.id,
    //         //   groupId: dbApiKey.groupId,
    //         //   userId: user.id,
    //         //   sessionId,
    //         //   version,
    //         // });
    //         //
    //         // // publish action to subscribers
    //         // pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
    //         //   faceCreated: { action: action1 },
    //         //   userId: dbApiKey.userId,
    //         //   groupId: dbApiKey.groupId,
    //         // });
    //         //
    //         // // const images = await utils.getFiles(files); // to_restore
    //         //
    //         // processFace({
    //         //   url: services.classify.recognition[version].url,
    //         //   images,
    //         //   models,
    //         //   sessionId,
    //         //   saveThumb: true,
    //         //   apiKey: dbApiKey,
    //         //   action: action1,
    //         //   resultPrefix: '',
    //         //   onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
    //         //     faceUpdated: { action: action1 },
    //         //     userId: dbApiKey.userId,
    //         //     groupId: dbApiKey.groupId,
    //         //   }),
    //         // });
    //         let action2;
    //
    //         for (let i = 0; i < 1; i++) {
    //           version = '2';
    //           sessionId = uuid4();
    //
    //           // add new action to db
    //           // const action2 = await models.FaceAction.create({
    //           action2 = await models.FaceAction.create({
    //             apiKeyId: dbApiKey.id,
    //             groupId: dbApiKey.groupId,
    //             userId: user.id,
    //             sessionId,
    //             version,
    //           });
    //
    //           // publish action to subscribers
    // console.log('Subscription.faceCreated!!!!!!!!!!!!!!!!!!!!!!!!!');
    //           pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED, {
    //             faceCreated: { action: action2 },
    //             userId: dbApiKey.userId,
    //             groupId: dbApiKey.groupId,
    //           });
    //
    //           // await utils.checkFiles(images, services.FACE_IMAGES);
    //           console.log(`\n${new Date()} [goto: processRecognition] ${sessionId}`);
    //           processRecognition({
    //             url: services.classify.recognition['2'].url,
    //             images,
    //             models,
    //             sessionId,
    //             saveThumb: true,
    //             apiKey: dbApiKey,
    //             action: action2,
    //             resultPrefix: '',
    //             onPublish: () => pubsub.publish(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED, {
    //               faceUpdated: { action: action2 },
    //               userId: dbApiKey.userId,
    //               groupId: dbApiKey.groupId,
    //             }),
    //           });
    //         }
    //
    // console.log('+++ RETURN ', action2);
    //         return action2;
    //       }
    //     ),

  },

  FaceAction: {
    user: async (action, args, { loaders }) => {
      return loaders.user.load(parseInt(action.userId, 10));
    },
    group: async (action, args, { loaders }) => {
      return loaders.group.load(parseInt(action.groupId, 10));
    },
    apiKey: async (action, args, { loaders }) => {
      return loaders.apiKey.load(parseInt(action.apiKeyId, 10));
    },
    person: async (item, args, { loaders }) => {
      console.log('item.personId', item.personId);
      const noPerson = { id: 0, thumb: '' };
      if (!item.personId) return noPerson;
      const person = await loaders.person.load(parseInt(item.personId, 10));
      // console.log('person result', person ? person : noPerson);
      return person || noPerson;
    },
    imageUrl: action => {
      if (action.result === null) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_thumb_50x50.jpg?${utils.getUnixTimeStamp(action.updatedAt)}`;
    },
    imageOriginalUrl: action => {
      if (action.result === null) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_0.jpg`;
    },
    result: (item, args, { me }) => {
      // console.log(' >>> GOTO classifyResult', item.result);
      const res = tools.classifyResult(item.result, item.createdAt, services.classify.recognition[item.version.toString()], isAdminUser(me));
      // console.log(' +++ FINALLY classifyResult', res);
      return res;
    },
  },

  Subscription: {
    faceCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.FACE_ACTIONS.FACE_ACTION_CREATED),
        (payload, variables, context) => {
          // console.log('Subscription.faceCreated', 'payload.userId', payload.userId, 'context.me.id', context.me.id);
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
          // return context.me.id === payload.userId && payload.groupId === parseInt(variables.groupId, 10);
        }
      ),
    },
    faceUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.FACE_ACTIONS.FACE_ACTION_UPDATED),
        (payload, variables, context) => {
          // console.log('Subscription.faceUpdated', 'payload.userId', payload.userId, 'context.me.id', context.me.id);
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
          // return context.me.id === payload.userId && payload.groupId === parseInt(variables.groupId, 10);
        }
      ),
    },
  },
};
