import uuid4 from 'uuid';
import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter } from 'apollo-server';
import * as utils from '../common/utils';
import { isAdminUser } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as constants from '../common/constants';
import * as services from '../services/constants';
import { processLiveness } from './serviceLiveness';
import { processFace, processRecognition } from "./serviceFaceAction";
import * as processes from '../services/processing';
import * as tools from '../services/tools';

export default {
  Query: {
    liveFaceAction: combineResolvers(
      async (parent, { id, token }, { models, me, secret }) => {
        const user = me || await utils.validateStatsToken(token, secret);
        const action = await models.LiveFaceAction.findById(id);
        if (!action) {
          throw new ValidationError(
            'No records found.'
          );
        }
        if (action.userId !== user.id && !isAdminUser(user)) {
          throw new ValidationError(
            'Access violation.'
          );
        }
        return action;
      }
    ),
    liveFaceActions: combineResolvers(
      async (parent, {
        groupId, token, cursor, limit = 100,
      }, { models, me, secret }) => {
        return processes.getServiceResults({
          user: me || await utils.validateStatsToken(token, secret),
          serviceModel: models.LiveFaceAction,
          groupId,
          cursor,
          limit,
          models,
        });
      }
    ),
  },

  Mutation: {

    goLiveFace: combineResolvers(
      async (root, {
        files, apiKey, versionLive, versionFace,
      }, { models, me }) => {
        utils.checkFiles(files, services.LIVE_FACE_IMAGES);

        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        const sessionId = uuid4();
        const isAdmin = isAdminUser(user);
        versionLive = tools.checkVersion('liveness', versionLive, isAdmin);
        versionFace = tools.checkVersion('recognition', versionFace, isAdmin);

        // add new action to db
        const action = await models.LiveFaceAction.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: dbApiKey.userId,
          sessionId,
          versionLive,
          versionFace,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_CREATED, {
          liveFaceCreated: { action },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        // await utils.checkFiles(images, services.LIVE_FACE_IMAGES);

        processLiveness({
          url: services.classify.liveness[versionLive].url,
          images,
          sessionId,
          action,
          resultPrefix: 'Live',
          onPublish: () => pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_UPDATED, {
            liveFaceUpdated: { action },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        processFace({
          url: services.classify.recognition['1'].url,
          images,
          models,
          sessionId,
          saveThumb: false,
          apiKey: dbApiKey,
          action,
          resultPrefix: 'Face',
          onPublish: () => pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_UPDATED, {
            liveFaceUpdated: { action },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        return action;
      }
    ),

    liveFace: combineResolvers(
      async (root, {
        files, apiKey, versionLive, versionFace,
      }, { models, me }) => {
        utils.checkFiles(files, services.LIVE_FACE_IMAGES);

        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        const sessionId = uuid4();
        const isAdmin = isAdminUser(user);
        versionLive = tools.checkVersion('liveness', versionLive, isAdmin);
        versionFace = tools.checkVersion('recognition', versionFace, isAdmin);

        // add new action to db
        const action = await models.LiveFaceAction.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: dbApiKey.userId,
          sessionId,
          versionLive,
          versionFace,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_CREATED, {
          liveFaceCreated: { action },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        // await utils.checkFiles(images, services.LIVE_FACE_IMAGES);

        await Promise.all([
          processLiveness({
            url: services.classify.liveness[versionLive].url,
            images,
            sessionId,
            action,
            resultPrefix: 'Live',
            onPublish: () => pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_UPDATED, {
              liveFaceUpdated: { action },
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            }),
          }),
          processRecognition({
            url: services.classify.recognition[2].url,
            images,
            models,
            sessionId,
            saveThumb: true,
            apiKey: dbApiKey,
            action,
            resultPrefix: 'Face',
            onPublish: () => pubsub.publish(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_UPDATED, {
              liveFaceUpdated: { action },
              userId: dbApiKey.userId,
              groupId: dbApiKey.groupId,
            }),
          }),
        ]);
        return action;
      }
    ),

  },

  LiveFaceAction: {
    user: async (action, args, { loaders }) => {
      return loaders.user.load(parseInt(action.userId, 10));
    },
    group: async (action, args, { loaders }) => {
      return loaders.group.load(parseInt(action.groupId, 10));
    },
    apiKey: async (action, args, { loaders }) => {
      return loaders.apiKey.load(parseInt(action.apiKeyId, 10));
    },
    person: (item, args, { loaders }) => {
      return item.personId
        ? loaders.person.load(parseInt(item.personId, 10))
        : { id: 0, thumb: '' };
    },
    imageUrl: action => {
      if (
        action.resultFace === null
        && action.resultLive === null
      ) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_thumb_50x50.jpg?${utils.getUnixTimeStamp(action.updatedAt)}`;
    },
    imageUrl200x200: action => {
      if (
        action.resultFace === null
        && action.resultLive === null
      ) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_thumb_200x200.jpg?${utils.getUnixTimeStamp(action.updatedAt)}`;
    },
    imageOriginalUrl: action => {
      if (
        action.resultFace === null
        && action.resultLive === null
      ) return '';
      const res = utils.createImagePath(action.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_0.jpg`;
    },
    resultLive: (item, args, { me }) => {
      return tools.classifyResult(
        item.resultLive,
        item.createdAt,
        services.classify.liveness[item.versionLive.toString()], isAdminUser(me)
      );
    },
    resultFace: (item, args, { me }) => {
      return tools.classifyResult(
        item.resultFace,
        item.createdAt,
        services.classify.recognition[item.versionFace.toString()], isAdminUser(me)
      );
    },
  },

  Subscription: {
    liveFaceCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_CREATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
        }
      ),
    },
    liveFaceUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.LIVE_FACE_ACTIONS.LIVE_FACE_ACTION_UPDATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
        }
      ),
    },
  },
};
