import uuid4 from 'uuid';
import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter } from 'apollo-server';
import * as utils from '../common/utils';
import { isAdminUser } from './authorization';
import pubsub, { EVENTS } from '../subscription';
import * as constants from '../common/constants';
import * as services from '../services/constants';
import * as processes from '../services/processing';
import * as tools from '../services/tools';

export const processLiveness = async ({
  url,
  images,
  sessionId,
  action,
  resultPrefix = '',
  onPublish,
}) => {
  // await for all processings done
  utils.processAWS({
    sessionId,
    images,
  });

  const [processingResult] = await Promise.all([
    processes.processLiveness({
      url,
      images,
    }),
  ]);

  utils.logAction(`processLiveness DONE, RESULT: ${processingResult}`);

  // publish action to subscribers
  await action.update({
    [`time${resultPrefix}`]: processingResult.processingTime,
    [`result${resultPrefix}`]: processingResult.value,
  });

  // save images to aws
  utils.processAWSThumb({
    imageBuffer: processingResult.thumb,
    sessionId,
    prefix: '',
    onUpdateAction: onPublish,
  });
  console.log('>>>>LIVENESS DONE', JSON.stringify(action));
  return action;
};

export default {
  Query: {
    livenessItem: combineResolvers(
      async (parent, { id, token }, { models, me, secret }) => {
        const user = me || await utils.validateStatsToken(token, secret);
        const action = await models.Liveness.findById(id);
        if (!action) {
          throw new ValidationError(
            'No records found.'
          );
        }
        if (action.userId !== user.id && !isAdminUser(user)) {
          throw new ValidationError(
            'Access violation.'
          );
        }
        return action;
      }
    ),
    livenessActions: combineResolvers(
      async (parent, {
        groupId, token, cursor, limit = 100,
      }, { models, me, secret }) => {
        return processes.getServiceResults({
          user: me || await utils.validateStatsToken(token, secret),
          serviceModel: models.Liveness,
          groupId,
          cursor,
          limit,
          models,
        });
      }
    ),
  },

  Mutation: {

    goLiveness: combineResolvers(
      async (root, { files, apiKey, version }, { models, me }) => {
        utils.checkFiles(files, services.LIVENESS_IMAGES);

        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        const sessionId = uuid4();
        version = tools.checkVersion('liveness', version, isAdminUser(user));

        // add new action to db
        const action = await models.Liveness.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: user.id,
          sessionId,
          version,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.LIVENESS.LIVENESS_CREATED, {
          livenessCreated: {
            action,
          },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        // utils.checkFiles(images, services.LIVENESS_IMAGES);
        // console.log('+++++++++++++++++++++++++++++++++++++UUUUUUUURRRRRRRLLLLLLLLLL', services.classify.liveness[version].url);
        processLiveness({
          url: services.classify.liveness[version].url,
          images,
          sessionId,
          action,
          onPublish: () => pubsub.publish(EVENTS.LIVENESS.LIVENESS_UPDATED, {
            livenessUpdated: { action },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });

        return action;
      }
    ),

    liveness: combineResolvers(
      async (root, { files, apiKey, version }, { models, me }) => {
        utils.checkFiles(files, services.LIVENESS_IMAGES);

        // preparations
        const { dbApiKey, user } = await utils.checkApiKey(apiKey, models, me);
        utils.updateApiKey(dbApiKey);
        const sessionId = uuid4();
        version = tools.checkVersion('liveness', version, isAdminUser(user));

        // add new action to db
        const action = await models.Liveness.create({
          apiKeyId: dbApiKey.id,
          groupId: dbApiKey.groupId,
          userId: user.id,
          sessionId,
          version,
        });

        // publish action to subscribers
        pubsub.publish(EVENTS.LIVENESS.LIVENESS_CREATED, {
          livenessCreated: {
            action,
          },
          userId: dbApiKey.userId,
          groupId: dbApiKey.groupId,
        });

        const images = await utils.getFiles(files);
        return processLiveness({
          url: services.classify.liveness[version].url,
          images,
          sessionId,
          action,
          onPublish: () => pubsub.publish(EVENTS.LIVENESS.LIVENESS_UPDATED, {
            livenessUpdated: { action },
            userId: dbApiKey.userId,
            groupId: dbApiKey.groupId,
          }),
        });
      }
    ),

  },

  Liveness: {
    user: (item, args, { loaders }) => {
      return loaders.user.load(parseInt(item.userId, 10));
    },
    group: (item, args, { loaders }) => {
      return loaders.group.load(parseInt(item.groupId, 10));
    },
    apiKey: (item, args, { loaders }) => {
      return loaders.apiKey.load(parseInt(item.apiKeyId, 10));
    },
    imageUrl: item => {
      if (item.result === null) return '';
      const res = utils.createImagePath(item.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_thumb_50x50.jpg?${utils.getUnixTimeStamp(item.updatedAt)}`;
    },
    imageOriginalUrl: item => {
      if (item.result === null) return '';
      const res = utils.createImagePath(item.sessionId);
      return `${constants.AWS_BUCKET_URL}${res.path}${res.name}_0.jpg`;
    },
    result: (item, args, { me }) => {
      const res = tools.classifyResult(item.result, item.createdAt, services.classify.liveness[item.version.toString()], isAdminUser(me));
      // console.log(' +++ FINALLY', res);
      return res;
    },
  },

  Subscription: {
    livenessCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.LIVENESS.LIVENESS_CREATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
        }
      ),
    },
    livenessUpdated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.LIVENESS.LIVENESS_UPDATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId
            && (!variables.groupId || (payload.groupId === parseInt(variables.groupId, 10)));
        }
      ),
    },
  },
};
