import Sequelize from 'sequelize';
import moment from 'moment';
import * as tools from '../services/tools';
import * as constants from '../services/constants';
import * as utils from '../common/utils';
import { isAdminUser } from './authorization';

const { Op } = Sequelize;

const getServiceModel = (id, models) => {
  switch (id) {
    case 1:
      return models.Liveness;
    case 2:
      return models.FaceAction;
    case 3:
      return models.Documents;
    case 4:
      return models.LiveFaceAction;
    case 5:
      return models.GuestAction;
    default:
      return null;
  }
};

const getServiceStats = async (service, models) => {
  let totalRequests = 0,
    successRequests = 0,
    failedRequests = 0;
  const serviceModel = getServiceModel(service.id, models);
  totalRequests = await serviceModel.count({
    col: 'id',
  });
  if (service.results.length === 1) {
    successRequests = await serviceModel.count({
      where: {
        result: {
          [Op.gt]: 50,
        },
      },
      col: 'id',
    });
  } else if (service.results.length === 2) {
    successRequests = await serviceModel.count({
      where: {
        resultLive: {
          [Op.gt]: 50,
        },
      },
      col: 'id',
    });
  }
  failedRequests = totalRequests - successRequests;
  return {
    id: service.id,
    name: service.name,
    totalRequests,
    successRequests,
    failedRequests,
  };
};

export default {
  Query: {
    services: (parent, args, { me }) => {
      let res = Object.values(
        constants.instances.filter(item => item.public)
      );
      res = res.map(item => {
        item.versions = item.results
          ? item.results.map(classificator => {
            return isAdminUser(me)
              ? Object.keys(classificator).filter(key => !['default'].includes(key))
              : [classificator.default];
          })
          : [];
        return item;
      });
      // console.log('FINAL SERVICES', res);
      return res;
    },
    service: (parent, { id }) => {
      return constants.instances[id];
    },
    servicesOverview: async (parent, args, { models }) => {
      return {
        services: await Promise.all(
          constants
            .instances
            .filter(item => item.public && item.withStats && item.results)
            .map(service => getServiceStats(service, models))
        ),
      };
    },
    serviceUsageByHour: async (parent, { id }, { models, me }) => {
      // console.log('[serviceUsageByHour] id', id, 'me', me);
      if (!me) return [];
      const service = tools.serviceById(parseInt(id, 10));
      // console.log('[serviceUsageByHour] service', service);
      const groupIds = await utils.prepareGroupIds({ models, groupId: null, user: me });
      // console.log('[serviceUsageByHour] groupIds', groupIds);
      const dbRes = await models.sequelize.query(
        `select date_trunc('hour', "${service.tableName}"."createdAt"), count(1)
          from "${service.tableName}"
          where "${service.tableName}"."createdAt" > '${moment.utc().subtract(23, 'hours').format('YYYY,MM-DD HH:mm:ss')}'
          and "${service.tableName}"."groupId" in (${groupIds})
          group by 1 order by 1`,
        { type: Sequelize.QueryTypes.SELECT }
      );
      console.log('[serviceUsageByHour] dbRes', dbRes);
      const intervals = utils.create24HourIntervals();
      for (const record of dbRes) {
        const hour = moment.utc(record.date_trunc).format('HH');
        for (let i = 0; i < intervals.length; i++) {
          if (hour === intervals[i].hour) {
            intervals[i].count = parseInt(record.count, 10);
            break;
          }
        }
      }
      return intervals
        .map((item, index) => ({ count: item.count, hourDelta: index }))
        .reverse();
    },
  },
};
