import { combineResolvers } from 'graphql-resolvers';
import {
  AuthenticationError, UserInputError, ValidationError, ForbiddenError,
} from 'apollo-server';
// import fs from 'fs';
import uuid4 from 'uuid';
import { isAdmin, isGod, isAuthenticated } from './authorization';
import * as utils from '../common/utils';
import * as constants from '../common/constants';
import * as services from '../services/constants';
import * as processes from '../services/processing';

export const createUser = async ({
  models,
  username,
  email,
  password,
  actionsLeft,
  role,
}) => {
  actionsLeft || (actionsLeft = constants.NEW_USER_API_KEYS_ACTIONS);
  role || (role = constants.USER_ROLE_CLIENT);
  const existingUser = await models.User.findByLogin(email);
  if (existingUser) return existingUser;
  const user = await models.User.create({
    username,
    email,
    password,
    role,
    uuid: uuid4(),
  });
  const group = await models.Group.create({
    userId: user.id,
    isActive: true,
    name: username,
  });
  await Promise.all(
    services
      .instances
      .filter(item => item.withApiKey)
      .map(service => {
        return models.ApiKey.create({
          state: constants.KEY_STATE_ACTIVE,
          type: constants.KEY_TYPE_DEFAULT,
          expired: 0,
          groupId: group.id,
          actionsLeft,
          serviceId: service.id,
          userId: user.id,
          uuid: uuid4(),
        });
      })
  );
  return user;
};

export default {
  Query: {
    user: combineResolvers(
      isAdmin,
      async (parent, { id }, { models }) => {
        return models.User.findById(id);
      }
    ),
    users: combineResolvers(
      isAdmin,
      async (parent, args, { models }) => {
        return models.User.findAll({
          order: [
            ['createdAt', 'DESC'],
          ],
        });
      }
    ),
    me: async (parent, args, { models, me }) => {
      if (!me || !me.id) return null;
      return models.User.findById(me.id);
    },
  },

  Mutation: {
    signUp: async (
      parent,
      { username, email, password },
      { models, secret }
    ) => {
      //      throw new ValidationError(
      //        'Registrations closed'
      //      );
      const existingUser = await models.User.findByLogin(email);
      if (existingUser) {
        throw new ValidationError(
          'User with this email address already exists'
        );
      }
      const user = await createUser({
        models, username, email, password,
      });
      return {
        token: await utils.createToken(user, secret, constants.SESSION_JWK_TIMEOUT),
        statsToken: await utils.createStatsToken(user, secret),
      };
    },

    signIn: async (
      parent,
      { email, password },
      { models, secret }
    ) => {
      const user = await models.User.findByLogin(email);
      if (!user) {
        throw new UserInputError(
          'No user found with this login credentials'
        );
      }
      if (!user.isActive) {
        throw new UserInputError(
          'This account is suspended'
        );
      }
      const isValid = await user.validatePassword(password);
      if (!isValid) {
        throw new AuthenticationError('Invalid password');
      }

      return {
        token: await utils.createToken(user, secret, constants.SESSION_JWK_TIMEOUT),
        statsToken: await utils.createStatsToken(user, secret),
      };
    },

    updateUser: combineResolvers(
      isAuthenticated,
      async (parent, { username, biometricSignIn }, { models, me }) => {
        utils.logAction(`
          updateUser
          username: ${username}
          biometricSignIn: ${biometricSignIn}
        `);
        const user = await models.User.findById(me.id);
        if (!user) {
          throw new ForbiddenError('Access denied');
        }
        return user.update({
          username, biometricSignIn,
        });
      }
    ),

    userSetActive: combineResolvers(
      isAdmin,
      async (parent, { id, isActive }, { models }) => {
        const user = await models.User.findById(id);
        if (!user) {
          throw new ForbiddenError('Access denied');
        }
        return user.update({
          isActive,
        });
      }
    ),

    userSetAdmin: combineResolvers(
      isGod,
      async (parent, { id, isAdministrator }, { models }) => {
        const user = await models.User.findById(id);
        if (!user) {
          throw new ForbiddenError('Access denied');
        }
        return user.update({
          role: isAdministrator ? constants.USER_ROLE_ADMIN : constants.USER_ROLE_CLIENT,
        });
      }
    ),

    goAvatar: combineResolvers(
      isAuthenticated,
      async (root, { files }, { models, me }) => {
        try {
          const images = await utils.getFiles(files);
          await utils.checkFiles(images, services.AVATAR_IMAGES);
          const myMainGroup = await models.Group.findByOwnerIdFirst(me.id);
          utils.logAction(`myMainGroup: ${myMainGroup.id}`);
          await processes.addPersonIdentix({
            models,
            images,
            userId: me.id,
            groupId: null,
            sourceId: null,
          });
        } catch (error) {
          throw new ValidationError(error);
        }
        return me;
      }
    ),

  },

  User: {
    avatar: async (user, args, { models }) => {
      const person = await models.Person.getLastChangedRecord(user.id);
      if (!person) return '';
      const AWSPath = utils.createImagePath(person.uuid);
      return `${constants.AWS_BUCKET_URL}${constants.AWS_PERSONS_PREFIX}${AWSPath.path}${AWSPath.name}_thumb_50x50.jpg?${Math.random()}`;
    },
  },
};
