import { combineResolvers } from 'graphql-resolvers';
import { ValidationError, withFilter, ForbiddenError } from 'apollo-server';
import moment from 'moment';
import pubsub, { EVENTS } from '../subscription';
import { isAdmin, isAuthenticated, isAdminUser } from './authorization';

export default {
  Query: {
    group: combineResolvers(
      isAdmin,
      async (parent, { id }, { models }) => {
        return models.Group.findById(id);
      }
    ),
    groups: combineResolvers(
      isAuthenticated,
      async (parent, args, { models, me }) => {
        return models.Group.findByOwnerId(me.id);
      }
    ),
    userGroups: combineResolvers(
      async (parent, { id }, { models, me }) => {
        if (id !== me.id && !isAdminUser(me)) {
          throw new ForbiddenError(
            'Access denied.'
          );
        }
        return models.Group.findByOwnerId(id);
      }
    ),
    groupMembers: combineResolvers(
      isAuthenticated,
      async (parent, { groupId }, { models }) => {
        return models.GroupMembers.findByGroupId(groupId);
      }
    ),
  },

  Mutation: {

    createGroup: combineResolvers(
      isAuthenticated,
      async (
        parent, { name }, { models, me }
      ) => {
        const group = await models.Group.create({
          userId: me.id,
          isActive: true,
          name: name || moment().format('MMM, DD YYYY HH:mm:ss'),
        });
        pubsub.publish(EVENTS.GROUPS.CREATED, {
          groupCreated: { group },
          userId: me.id,
        });
        return group;
      }
    ),

    updateGroup: combineResolvers(
      isAuthenticated,
      async (
        parent, { id, isActive, name }, { models, me }
      ) => {
        const group = await models.Group.findById(id);
        if (!group) {
          throw new ValidationError(
            'No group found with this id.'
          );
        }
        if (group.userId !== me.id && !isAdminUser(me)) {
          throw new ValidationError(
            'This is not your group.'
          );
        }
        if (!isActive) {
          // request for group deletion, additional checks
          const mainGroup =  await models.Group.findByOwnerIdFirst(me.id);
          if (mainGroup.get({ plain: true }).id === parseInt(id, 10)) {
            throw new ValidationError(
              'Can not remove default user group.'
            );
          }
        }
        return group.update({
          isActive,
          name: name || group.name,
        });
      }
    ),

    joinGroup: combineResolvers(
      isAuthenticated,
      async (
        parent, { email }, { models }
      ) => {
        const user = await models.User.findByLogin(email);
        if (!user) {
          throw new ValidationError(
            'User not found.'
          );
        }
        const userMainGroup = await models.Group.findByOwnerIdFirst(user.id);
        if (!userMainGroup) {
          throw new ValidationError(
            'User has no groups.'
          );
        }
        const groupMember = await models.GroupMembers.findByGroupIdUserId(userMainGroup.id, user.id);
        if (groupMember) {
          throw new ValidationError(
            'You request is under user review.'
          );
        }
      }
    ),

    addMember: combineResolvers(
      isAuthenticated,
      async (
        parent, { groupId, userId, role }, { models }
      ) => {
        const group = await models.Group.findById(groupId);
        if (!group) {
          throw new ValidationError(
            'No group found with thi id.'
          );
        }
        const user = await models.GroupMembers.findByGroupIdUserId(groupId, userId);
        if (user) {
          throw new ValidationError(
            'User is already in this group.'
          );
        }
        return models.GroupMembers.create({
          groupId,
          userId,
          role,
        });
      }
    ),

    removeMember: combineResolvers(
      isAuthenticated,
      async (
        parent, { groupId, userId }, { models }
      ) => {
        const group = await models.Group.findById(groupId);
        if (!group) {
          throw new ValidationError(
            'No group found with thi id.'
          );
        }
        const user = await models.GroupMembers.findByGroupIdUserId(groupId, userId);
        if (user) {
          throw new ValidationError(
            'User is already in this group.'
          );
        }
        return models.GroupMembers.destroy({
          where: {
            groupId,
            userId,
          },
        });
      }
    ),

    updateMember: combineResolvers(
      isAuthenticated,
      async (
        parent, { groupId, userId, role }, { models }
      ) => {
        const group = await models.Group.findById(groupId);
        if (!group) {
          throw new ValidationError(
            'No group found with thi id.'
          );
        }
        const user = await models.GroupMembers.findByGroupIdUserId(groupId, userId);
        if (user) {
          throw new ValidationError(
            'User is already in this group.'
          );
        }
        return group.update({
          groupId,
          userId,
          role,
        });
      }
    ),

  },

  Group: {
  },

  Subscription: {
    groupCreated: {
      subscribe: withFilter(
        () => pubsub.asyncIterator(EVENTS.GROUPS.CREATED),
        (payload, variables, context) => {
          return context.me.id === payload.userId;
        }
      ),
    },
  },
};
