import Sequelize from 'sequelize';
import uuid4 from 'uuid';
import axios from 'axios/index';
import FormData from 'form-data';
import { ValidationError } from 'apollo-server/dist/index';
import WebSocket from 'ws';
import * as services from './constants';
import * as tools from './tools';
import * as utils from '../common/utils';
import * as constants from '../common/constants';

const { Op } = Sequelize;

export const processLiveness = async ({
  url,
  images,
}) => {
  const result = {
    value: null,
    thumb: images[0],
    processingTime: null,
  };
  utils.logAction(`
    processLiveness
    Total images: ${images.length}
    Image sizes: ${images.map(item => item.byteLength)}
  `);
  return new Promise(async resolve => {
    try {
      const form = new FormData();
      for (let i = 0; i < images.length; i++) {
        form.append('upload[]', images[i], `${i}.jpg`);
      }
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: 30000,
        transformRequest: [(data, headers) => {
          headers.post['User-Agent'] = 'faucet.js';
          return data;
        }],
      });
      utils.logAction(`liveness url: ${url}`);
      const response = await axiosInstance.post(url, form);
      if (
        !response.data
        || !response.data.result
        || !Array.isArray(response.data.result)
        || !response.data.result.length
      ) {
        result.value = services.SERVICE_ERROR_WRONG_RESPONSE;
      } else {
        result.value = response.data.result[0] >= 0
          ? Math.round(response.data.result[0] * services.RESULT_PRECISION) / services.RESULT_PRECISION
          : response.data.result[0];
      }
      result.thumb = response.data.file ? Buffer.from(response.data.file, 'base64') : images[0];
      result.processingTime = response.data.time;
    } catch (error) {
      result.value = tools.classifyError(error);
    } finally {
      resolve(result);
    }
  });
};

export const processFaceCompare = async ({
  images,
  compareWith,
}) => {
  const result = {
    value: null,
    thumb: images[0],
    processingTime: null,
  };
  return new Promise(async resolve => {
    try {
      const form = new FormData();
      form.append('photo1', compareWith, '1.jpg');
      form.append('photo2', images[0], '2.jpg');
      form.append('liveness_photo1', 'true');
      form.append('liveness_photo2', 'true');
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: 30000,
        transformRequest: [(data, headers) => {
          headers.post.Authorization = `Token ${services.IDENTIX_TOKEN}`;
          return data;
        }],
      });
      const response = await axiosInstance.post(services.FACE_URL_COMPARE, form);
      result.value = utils.identixResponseToInt(response.data.conf);
    } catch (error) {
      result.value = tools.classifyError(error);
    } finally {
      resolve(result);
    }
  });
};

export const processFaceCreate = async ({
  images,
}) => {
  const result = {
    value: null,
    thumb: images[0],
    processingTime: null,
    idxid: null,
  };
  return new Promise(async resolve => {
    try {
      const form = new FormData();
      form.append('photo', images[0], '1.jpg');
      form.append('source', 'api-source');
      form.append('facesize', 300);
      form.append('create_on_ha', 'false');
      form.append('create_on_junk', 'true');
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: 30000,
        transformRequest: [(data, headers) => {
          headers.post.Authorization = `Token ${services.IDENTIX_TOKEN}`;
          return data;
        }],
      });
      const response = await axiosInstance.post(services.FACE_URL_CREATE, form);
      if (
        !response.data
        || !response.data.idxid
      ) {
        result.value = services.SERVICE_ERROR_WRONG_RESPONSE;
      } else {
        result.value = tools.identixResponseToInt(response.data.conf);
        result.idxid = response.data.idxid;
      }
    } catch (error) {
      result.value = tools.classifyErrorIdentix(error);
    } finally {
      resolve(result);
    }
  });
};

export const processFaceFind = async ({
  url,
  images,
}) => {
  const result = {
    value: null,
    thumb: images[0],
    processingTime: null,
    uuid: null,
  };
  return new Promise(async resolve => {
    try {
      const form = new FormData();
      form.append('photo', images[0], '1.jpg');
      form.append('liveness', 'true');
      form.append('asm', 'true');
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: services.SERVICE_DELAY_TO_FAIL * 1000,
        transformRequest: [(data, headers) => {
          headers.post.Authorization = `Token ${services.IDENTIX_TOKEN}`;
          return data;
        }],
      });
      const response = await axiosInstance.post(url, form);
      utils.logAction(`IDENTIX RESPONSE ${JSON.stringify(response.data)}`);
      if (
        !response.data
        || !response.data.idxid
        || !response.data.conf
      ) {
        result.value = services.SERVICE_ERROR_WRONG_RESPONSE;
        return;
      }
      result.value = tools.identixResponseToInt(response.data.conf);
      result.uuid = response.data.idxid;
    } catch (error) {
      result.value = tools.classifyErrorIdentix(error);
    } finally {
      resolve(result);
    }
  });
};

export const processRecognition = async ({
  userId,
  groupId,
  createDistance,
  newPersonUuid,
  url,
  images,
  sessionId, // to_del
}) => {
  console.log(`[processRecognition] ${sessionId}`);
  const result = {
    value: null,
    thumb: images[0],
    processingTime: null,
    uuid: null,
  };
  return new Promise(async resolve => {
    try {
      const form = new FormData();
      form.append('upload', images[0], '1.jpg');
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: services.SERVICE_DELAY_TO_FAIL * 1000,
        transformRequest: [(data, headers) => {
          headers.post['x-user-id'] = userId;
          groupId === undefined || (headers.post['x-group-id'] = groupId);
          if (
            createDistance !== undefined
            && newPersonUuid !== undefined
          ) {
            headers.post['x-create-distance'] = createDistance;
            headers.post['x-new-person-uuid'] = newPersonUuid;
          }
          return data;
        }],
      });
      const response = await axiosInstance.post(url, form);
      console.log(`ENFACE RESPONSE  ${sessionId} ${JSON.stringify(response.data)}`);
      if (
        !response.data
        || response.data.result === undefined
        || response.data.time === undefined
        || response.data.uuid === undefined
        // || !isUUID(response.data.uuid)
      ) {
        // result.value = services.SERVICE_ERROR_WRONG_RESPONSE;
        console.log('BAD RESPONSE');
        // result.value = services.classify.recognition['2'].scoring[2].failed_strong;
        result.value = services.SERVICE_ERROR_SERVICE;
        return;
      }
      // && (response.data.result = [services.classify.recognition['2'].scoring.length - 1]);
      result.value = response.data.result >= 0
        ? Math.round(response.data.result * services.RESULT_PRECISION) / services.RESULT_PRECISION
        : response.data.result;
      if (response.data.result === services.SERVICE_ERROR_NO_USER_ID) {
        const classificator = services.classify.recognition['2'].scoring;
        // utils.logAction('ENFACE ERROR NO USER ID, classificator', classificator);
        // utils.logAction('ENFACE ERROR NO USER ID, key', Object.keys(classificator[classificator.length - 1])[0]);
        result.value = classificator[classificator.length - 1][Object.keys(classificator[classificator.length - 1])[0]];
        // utils.logAction('ENFACE ERROR NO USER ID, RESULT', result.value);
      } else if (response.data.result === services.SERVICE_ERROR_NO_GROUP_ID) {
        const classificator = services.classify.recognition['2'].scoring;
        result.value = classificator[classificator.length - 1][Object.keys(classificator[classificator.length - 1])[0]];
        // utils.logAction('ENFACE ERROR NO GROUP ID, RESULT', result.value);
      }
      // utils.logAction('ENFACE RESULT', result.value, typeof result.value);
      result.uuid = response.data.uuid;
      result.processingTime = response.data.time;
    } catch (error) {
      utils.logAction(`ENFACE ERROR ${error}`);
      result.value = tools.classifyErrorRecognition(error);
    } finally {
      resolve(result);
    }
  });
};

export const addPersonIdentix = async ({
  models,
  images,
  groupId,
  userId,
  sourceId,
}) => {
  let person = null;
  return new Promise(async (resolve, reject) => {
    try {
      utils.logAction(`
    addPersonIdentix
    Total images: ${images.length}
    Image sizes: ${images.map(item => item.byteLength)}
    groupId: ${groupId}
    userId: ${userId}
      `);

      // await for images registered
      const addResult = await processFaceCreate({ images });
      if (addResult.value <= 0) reject(new Error(addResult.value));
      person = await models.Person.findWithParams({
        uuid: addResult.idxid,
        groupId,
        userId,
      });
      if (person) {
        // tricky
        await person.update({
          uuid: uuid4(),
          where: { groupId },
        });
        await person.update({
          uuid: addResult.idxid,
          where: { groupId },
        });
      } else {
        person = await models.Person.create({
          uuid: addResult.idxid,
          groupId,
          userId,
          sourceId,
        });
      }
      await Promise.all([
        utils.processAWS({
          images,
          sessionId: person.uuid,
          prefix: constants.AWS_PERSONS_PREFIX,
        }),
        utils.processAWSThumb({
          imageBuffer: images[0],
          sessionId: person.uuid,
          prefix: constants.AWS_PERSONS_PREFIX,
        }),
      ]);
      resolve(person);
    } catch (error) {
      reject(error);
    }
  });
};

export const addPerson = async ({
  models,
  images,
  bcId,
  userId,
  groupId,
  sourceId,
  createDistance,
  newPersonUuid,
  url,
}) => {
  let person = null;
  return new Promise(async (resolve, reject) => {
    try {
      utils.logAction(`
    addPerson
    Total images: ${images.length}
    Image sizes: ${images.map(item => item.byteLength)}
    groupId: ${groupId}
    userId: ${userId}
    sourceId: ${sourceId}
    createDistance: ${createDistance}
    newPersonUuid: ${newPersonUuid}
    url: ${url}
      `);
      // console.log('!!!!!!!!!!!!!!!!!!!!!!!!)))))))))))))))))' , typeof isUUID.anyNonNil);
      // return;
      const form = new FormData();
      form.append('upload', images[0], '1.jpg');
      const axiosInstance = axios.create({
        headers: form.getHeaders(),
        timeout: services.SERVICE_DELAY_TO_FAIL * 1000,
        transformRequest: [(data, headers) => {
          headers.post['x-user-id'] = userId;
          groupId === undefined || (headers.post['x-group-id'] = groupId);
          if (
            createDistance !== undefined
            && newPersonUuid !== undefined
          ) {
            headers.post['x-create-distance'] = createDistance;
            headers.post['x-new-person-uuid'] = newPersonUuid;
          }
          return data;
        }],
      });
      const response = await axiosInstance.post(url, form);
      utils.logAction(`ADD PERSON RESPONSE ${JSON.stringify(response.data)}`);
      if (
        !response.data
        || response.data.result === undefined
        || response.data.time === undefined
        || response.data.uuid === undefined
        || !utils.isUuid(response.data.uuid)
      ) {
        reject(new ValidationError('Wrong service response!'));
      }
      // if (
      //   response.data.result !== 0
      // ) {
      //   reject(new ValidationError('Service response is not zero!'));
      // }
      const where = {
        uuid: response.data.uuid,
        groupId,
        userId,
      };
      bcId && (where.bcId = bcId);
      person = await models.Person.findWithParams(where);
      if (person) {
        // tricky
        await person.update({
          uuid: uuid4(),
          where: { groupId },
        });
        await person.update({
          uuid: response.data.uuid,
          where: { groupId },
        });
      } else {
        person = await models.Person.create({
          uuid: response.data.uuid,
          groupId,
          userId,
          sourceId,
        });
      }
      await Promise.all([
        utils.processAWS({
          images,
          sessionId: response.data.uuid,
          prefix: constants.AWS_PERSONS_PREFIX,
        }),
        utils.processAWSThumb({
          imageBuffer: images[0],
          sessionId: response.data.uuid,
          prefix: constants.AWS_PERSONS_PREFIX,
        }),
      ]);
      resolve(person);
    } catch (error) {
      reject(error);
    }
  });
};

export const getServiceResults = async ({
  groupId,
  cursor,
  limit,
  models,
  serviceModel,
  user,
}) => {
  const groupIds = await utils.prepareGroupIds({ models, groupId, user });
  const cursorOptions = cursor
    ? {
      where: {
        id: {
          [Sequelize.Op.lt]: utils.fromCursorHash(cursor),
        },
      },
    }
    : {};
  cursorOptions.where = { ...cursorOptions.where, groupId: { [Op.in]: groupIds } };
  const result = await serviceModel.findAll({
    order: [
      ['createdAt', 'DESC'],
    ],
    limit: limit + 1,
    ...cursorOptions,
  });
  const hasNextPage = result.length > limit;
  const edges = hasNextPage ? result.slice(0, -1) : result;
  return {
    edges,
    pageInfo: {
      hasNextPage,
      endCursor: utils.toCursorHash(edges.length ? edges[edges.length - 1].id.toString() : ''),
    },
  };
};

export const sendAxios = async (url, data) => {
  try {
    utils.logAction(`checkBackendReady AXIOS url: ${url}`);
    const response = await axios({
      method: 'post',
      url,
      headers: {},
      data,
    });
    utils.logAction(`AXIOS RESPONSE: ${JSON.stringify(response.data)}`);
    return response.data;
  } catch (error) {
    console.log('AXIOS ERROR', error.message);
    return false;
  }
};

export const checkBackendReady = async ({
  url,
  sessionId,
}) => {
  console.log('checkBackendReady, url, sessionId', url, sessionId);
  let ws;
  const isHttp = url.startsWith('http');
  const terminate = listener => {
    isHttp || (listener && listener.terminate());
  };
  return new Promise(async (resolve, reject) => {
    const checkResponse = data => {
      switch (data._) {
        case constants.SOCKET_COMMAND_ERROR:
          console.log('Project backend response: ERROR', data.message);
          terminate(ws);
          reject(new Error(`Error project backend response: ${data.message}`));
          break;
        case constants.SOCKET_COMMAND_READY:
          console.log('Project backend response: READY');
          resolve(ws);
          break;
        default:
          console.log('Project backend response: UNKNOWN');
          terminate(ws);
          reject(new Error('Unknown project backend response'));
          break;
      }
    };
    const postData = {
      _: constants.SOCKET_COMMAND_CHECK,
      sessionId,
    };
    if (isHttp) {
      const result = await sendAxios(`${url}`, postData);
      result
        ? checkResponse(result)
        : reject(new Error('Connection failed'));
    } else {
      try {
        ws = new WebSocket(url, {
          handshakeTimeout: 10000,
        });
        ws.on('error', error => {
          console.log('[Socket.error]', error);
          reject(new Error('Connection failed'));
        });
        ws.on('open', () => {
          console.log('[Socket.opened]');
          ws.send(JSON.stringify(postData));
        });
        ws.on('close', (code, reason) => {
          console.log('[Socket.close]', code, reason);
          reject(new Error('socket closed'));
        });
        ws.on('message', message => {
          console.log('[Socket.message]', message);
          const data = JSON.parse(message);
          console.log('Received command', data._);
          checkResponse(data);
        });
      } catch (error) {
        reject(error.message);
        terminate(ws);
      }
    }
  });
};
