import moment from 'moment';
import * as constants from './constants';

export const getActionTime = (started, finished) => {
  return moment.utc(finished).diff(moment.utc(started)) / 1000;
};

export const getProcessingTime = processingTime => {
  return Math.round(processingTime * 1000) / 1000;
};

export const classifyResult = (result, createdAt, schema, withScore) => {
  // console.error('[classifyResult] schema', schema, withScore);
  // console.error('[classifyResult] item', result);
  const res = {
    success: false,
    confidence: '',
    error: '',
    score: 0,
  };
  if (!schema) {
    res.error = 'Bad version';
    return res;
  }
  // res.score = withScore ? result : 0;
  res.score = result;
  switch (result) {
    case null:
    case undefined:
      if (moment().diff(moment(createdAt)) > constants.SERVICE_DELAY_TO_FAIL * 1000) {
        res.error = 'service down';
      } else {
        res.success = null;
      }
      return res;
    case constants.SERVICE_ERROR_SERVICE:
      res.error = 'service error';
      return res;
    case constants.SERVICE_ERROR_NETWORK:
      res.error = 'network error';
      return res;
    case constants.SERVICE_ERROR_SERVER:
      res.error = 'server error';
      return res;
    case constants.SERVICE_ERROR_WRONG_RESPONSE:
      res.error = 'bad response';
      return res;
    case constants.SERVICE_ERROR_BAD_IMAGE:
      res.error = 'bad image';
      return res;
    case constants.SERVICE_ERROR_NO_FACE:
      res.error = 'no face';
      return res;
    case constants.SERVICE_ERROR_TOO_FAR:
      res.error = 'face is too far';
      return res;
    case constants.SERVICE_ERROR_TOO_CLOSE:
      res.error = 'face is too close';
      return res;
    default:
      break;
  }
  // console.log('  +++ goto scoring', result);
  for (const score of schema.scoring) {
    const key = Object.keys(score)[0];
    // console.log('  +++ next key', key, 'max score', score[key]);
    if (result <= score[key]) {
      // console.log('  +++ this value is result scoring');
      switch (key) {
        case 'success_strong':
          // console.log('  +++ this is success_strong!');
          res.success = true;
          res.confidence = 'strong';
          return res;
        case 'success_unsure':
          // console.log('  +++ this is success_unsure!');
          res.success = true;
          res.confidence = 'unsure';
          return res;
        case 'success_poor':
          // console.log('  +++ this is success_poor!');
          res.success = true;
          res.confidence = 'poor';
          return res;
        case 'failed_poor':
          res.confidence = 'poor';
          return res;
        case 'failed_unsure':
          res.confidence = 'unsure';
          return res;
        case 'failed_strong':
          res.confidence = 'strong';
          return res;
        default:
          break;
      }
    }
  }
  // console.log('  +++ this is success_strong without case!');
  res.success = true;
  res.confidence = 'strong';
  return res;
};

export const serviceById = id => {
  for (let i = 0; i < constants.instances.length; i++) {
    if (constants.instances[i].id === id) return constants.instances[i];
  }
  return null;
};

export const identixResponseToInt = match => {
  switch (match) {
    case 'exact':
      return constants.IDENTIX_EXACT;
    case 'new':
      return constants.IDENTIX_EXACT;
    case 'ha':
      return constants.IDENTIX_HA;
    case 'junk':
      return constants.IDENTIX_JUNK;
    case 'nm':
      return constants.IDENTIX_NM;
    case 'det':
      return constants.SERVICE_ERROR_TOO_FAR;
    default:
      return -1;
  }
};

export const classifyError = error => {
  if (error.response) {
    console.error('[Response.error]', error.response.status);
    return constants.SERVICE_ERROR_SERVICE;
  }
  if (error.request) {
    console.error('[Request.error]');
    return constants.SERVICE_ERROR_NETWORK;
  }
  console.error('[Internal.error]', error);
  return constants.SERVICE_ERROR_SERVER;
};

export const classifyErrorIdentix = (error) => {
  if (!error.response) {
    if (error.request) {
      console.error('[Request.error]');
      return constants.SERVICE_ERROR_NETWORK;
    }
    console.error('[Internal.error]', error);
    return constants.SERVICE_ERROR_SERVER;
  }
  if (error.response.status === 404) return 0;
  if (error.response.status === 400) {
    console.error('[Response.error]', error.response.status, error.response.data);
    if (
      error.response.data.photo
      && Array.isArray(error.response.data.photo)
      && error.response.data.photo.length
      && error.response.data.photo[0].indexOf('Upload a valid image') >= 0
    ) {
      return constants.SERVICE_ERROR_BAD_IMAGE;
    }
    if (error.response.data.detail.indexOf('Invalid image') >= 0) {
      return constants.SERVICE_ERROR_BAD_IMAGE;
    }
    if (error.response.data.detail.indexOf('Liveness check failed') >= 0) {
      return constants.SERVICE_ERROR_LIVENESS;
    }
    if (error.response.data.detail.indexOf('Unable to detect any faces on image') >= 0) {
      return constants.SERVICE_ERROR_NO_FACE;
    }
    if (error.response.data.detail.indexOf('Create source with this name') >= 0) {
      return constants.SERVICE_ERROR_NO_SOURCE;
    }
    return constants.SERVICE_ERROR_WRONG_RESPONSE;
  }
  console.error('[SERVICE_ERROR_UNKNOWN]', error);
  return constants.SERVICE_ERROR_UNKNOWN;
};

export const classifyErrorRecognition = (error) => {
  if (!error.response) {
    if (error.request) {
      console.error('[Request.error]');
      return constants.SERVICE_ERROR_NETWORK;
    }
    console.error('[Internal.error]', error);
    return constants.SERVICE_ERROR_SERVER;
  }
  if (error.response.status === 500) {
    return constants.SERVICE_ERROR_SERVICE;
  }
  return constants.SERVICE_ERROR_UNKNOWN;
};

export const checkVersion = (service, version, canUseNotDefault) => {
  // console.log('checkVersion!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!');
  // version = '1.01';
  if (
    (
      version !== constants.classify[service].default
      && !canUseNotDefault
    )
    || !version
    || !constants.classify[service][version]
  ) {
    version = constants.classify[service].default;
    console.log('WRONG VERSION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!');
  }
  console.log('FINAL VERSION - ' + service, version);
  return parseFloat(version);
};
